"""Module for staged resource error tracking."""

from sqlalchemy import Column, DateTime, ForeignKey, String

from fides.api.db.base_class import Base, FidesBase


class StagedResourceError(Base):
    """
    DB model for tracking errors associated with staged resources.
    Linked to StagedResource via FK with cascade delete.
    """

    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)
    staged_resource_urn = Column(
        String,
        ForeignKey("stagedresource.urn", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    message = Column(String, nullable=False)
    timestamp = Column(DateTime(timezone=True), nullable=False)
    error_type = Column(String, nullable=False)
    diff_status = Column(String, nullable=False)  # Phase where error occurred
    task_id = Column(String, nullable=True, index=True)



# pylint: disable=too-many-lines
from datetime import datetime
from typing import Any, Dict, List, Optional

from loguru import logger
from sqlalchemy.orm import Session, selectinload

from fides.api.common_exceptions import (
    FidesopsException,
    MessageDispatchException,
    PrivacyRequestError,
    RedisNotConfigured,
)
from fides.api.models.audit_log import AuditLog, AuditLogAction
from fides.api.models.policy import Policy
from fides.api.models.pre_approval_webhook import PreApprovalWebhook
from fides.api.models.privacy_center_config import (
    PrivacyCenterConfig as PrivacyCenterConfigModel,
)
from fides.api.models.privacy_preference import PrivacyPreferenceHistory
from fides.api.models.privacy_request import (
    ConsentRequest,
    ExecutionLog,
    PrivacyRequest,
    RequestTask,
)
from fides.api.models.property import Property
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.api import BulkUpdateFailed
from fides.api.schemas.messaging.messaging import MessagingActionType
from fides.api.schemas.policy import ActionType, CurrentStep
from fides.api.schemas.privacy_center_config import LocationCustomPrivacyRequestField
from fides.api.schemas.privacy_center_config import (
    PrivacyCenterConfig as PrivacyCenterConfigSchema,
)
from fides.api.schemas.privacy_request import (
    BULK_PRIVACY_REQUEST_BATCH_SIZE,
    BulkPostPrivacyRequests,
    BulkReviewResponse,
    CheckpointActionRequired,
    PrivacyRequestCreate,
    PrivacyRequestResponse,
    PrivacyRequestResubmit,
    PrivacyRequestSource,
    PrivacyRequestStatus,
)
from fides.api.service.messaging.message_dispatch_service import message_send_enabled
from fides.api.service.privacy_request.duplication_detection import check_for_duplicates
from fides.api.service.privacy_request.request_service import (
    build_required_privacy_request_kwargs,
    cache_data,
)
from fides.api.tasks import DSR_QUEUE_NAME
from fides.api.util.cache import cache_task_tracking_key
from fides.api.util.logger_context_utils import LoggerContextKeys, log_context
from fides.config.config_proxy import ConfigProxy
from fides.service.messaging.messaging_service import (
    MessagingService,
    check_and_dispatch_error_notifications,
    send_privacy_request_receipt_message_to_user,
)


class PrivacyRequestService:
    def __init__(
        self,
        db: Session,
        config_proxy: ConfigProxy,
        messaging_service: MessagingService,
    ):
        self.db = db
        self.config_proxy = config_proxy
        self.messaging_service = messaging_service

    def get_privacy_request(self, privacy_request_id: str) -> Optional[PrivacyRequest]:
        privacy_request: Optional[PrivacyRequest] = (
            PrivacyRequest.query_without_large_columns(self.db)
            .filter(PrivacyRequest.id == privacy_request_id)
            .first()
        )
        if not privacy_request:
            logger.info(f"Privacy request with ID {privacy_request_id} was not found.")
        return privacy_request

    def _validate_privacy_request_for_bulk_operation(
        self,
        privacy_request: Optional[PrivacyRequest],
        request_id: str,
    ) -> PrivacyRequest:
        """
        Common validation logic for bulk operations.
        Raises a PrivacyRequestError if the request is not found or deleted.

        Returns the validated privacy request.
        """
        if not privacy_request:
            raise PrivacyRequestError(
                f"No privacy request found with id '{request_id}'",
                {"privacy_request_id": request_id},
            )
        if privacy_request.deleted_at is not None:
            raise PrivacyRequestError(
                "Cannot transition status for a deleted request",
                PrivacyRequestResponse.model_validate(privacy_request).model_dump(
                    mode="json"
                ),
            )
        return privacy_request

    def _fetch_privacy_requests_for_bulk_operation(
        self, request_ids: List[str]
    ) -> Dict[str, PrivacyRequest]:
        """Fetch privacy requests in a single query to avoid N+1. Eager loads custom_fields."""
        privacy_requests = (
            PrivacyRequest.query_without_large_columns(self.db)
            .options(selectinload(PrivacyRequest.custom_fields))  # type: ignore[attr-defined]
            .filter(PrivacyRequest.id.in_(request_ids))
            .all()
        )
        return {pr.id: pr for pr in privacy_requests}

    def _validate_required_location_fields(
        self, privacy_request_data: PrivacyRequestCreate
    ) -> None:
        """Validate that location is provided for required location fields.

        Looks up the actual Privacy Center configuration to check if any location
        fields are marked as required for the specified policy.
        """
        # If location is already provided, no validation needed
        if privacy_request_data.location:
            return

        config_dict = self._resolve_privacy_center_config_dict(
            privacy_request_data.property_id
        )
        if not config_dict:
            return

        privacy_center_config = self._parse_privacy_center_config(config_dict)
        if not privacy_center_config:
            return

        action = self._get_matching_action(
            privacy_center_config, privacy_request_data.policy_key
        )
        if not action or not action.custom_privacy_request_fields:
            return

        missing_required = self._is_required_location_missing(
            action, privacy_request_data
        )
        if missing_required:
            # Find the first location field name to reference in the error for clarity
            first_location_field = next(
                (
                    name
                    for name, cfg in action.custom_privacy_request_fields.items()
                    if isinstance(cfg, LocationCustomPrivacyRequestField)
                    and getattr(cfg, "required", False)
                ),
                "location",
            )
            raise PrivacyRequestError(
                f"Location is required for field '{first_location_field}' but was not provided",
                privacy_request_data.model_dump(mode="json"),
            )

    def _resolve_privacy_center_config_dict(
        self, property_id: Optional[str]
    ) -> Optional[Dict[str, Any]]:
        """Return the applicable Privacy Center config as a dict using precedence.

        Precedence:
        1) The request's property config (if provided)
        2) The default property's config (if available)
        3) The single-row Privacy Center config table (legacy/global)
        """
        # 1) Request's property config
        if property_id:
            prop = Property.get_by(self.db, field="id", value=property_id)
            if prop and getattr(prop, "privacy_center_config", None):
                return prop.privacy_center_config  # type: ignore[return-value]

        # 2) Default property config
        default_prop = Property.get_by(self.db, field="is_default", value=True)
        if default_prop and getattr(default_prop, "privacy_center_config", None):
            return default_prop.privacy_center_config  # type: ignore[return-value]

        # 3) Single-row global config
        privacy_center_config_record = PrivacyCenterConfigModel.filter(
            db=self.db, conditions=PrivacyCenterConfigModel.single_row  # type: ignore[arg-type]
        ).first()
        if privacy_center_config_record:
            return privacy_center_config_record.config  # type: ignore[return-value]
        return None

    def _parse_privacy_center_config(
        self, config_dict: Dict[str, Any]
    ) -> Optional[PrivacyCenterConfigSchema]:
        """Parse a config dict into a schema instance, logging and skipping on error."""
        try:
            return PrivacyCenterConfigSchema.model_validate(config_dict)
        except Exception as exc:  # pylint: disable=broad-except
            logger.warning(
                f"Could not parse Privacy Center config for location validation: {exc}"
            )
            return None

    @staticmethod
    def _get_matching_action(
        privacy_center_config: PrivacyCenterConfigSchema, policy_key: str
    ) -> Optional[Any]:
        """Return the action entry matching the given policy key, if any."""
        for action in privacy_center_config.actions:
            if action.policy_key == policy_key:
                return action
        return None

    @staticmethod
    def _is_required_location_missing(
        action: Any, privacy_request_data: PrivacyRequestCreate
    ) -> bool:
        """Check if any required location fields exist without values in the request."""
        if not getattr(action, "custom_privacy_request_fields", None):
            return False

        custom_fields = action.custom_privacy_request_fields
        for field_name, field_config in custom_fields.items():
            if not isinstance(field_config, LocationCustomPrivacyRequestField):
                continue
            if not getattr(field_config, "required", False):
                continue

            # Check if this location field has a value in the request custom fields
            field_has_value = False
            if privacy_request_data.custom_privacy_request_fields:
                field_data = privacy_request_data.custom_privacy_request_fields.get(
                    field_name
                )
                if field_data and getattr(field_data, "value", None):
                    field_has_value = True

            if not field_has_value:
                return True

        return False

    # pylint: disable=too-many-branches, too-many-statements
    def create_privacy_request(
        self,
        privacy_request_data: PrivacyRequestCreate,
        *,
        authenticated: bool = False,
        privacy_preferences: Optional[List[PrivacyPreferenceHistory]] = None,
        submitted_by: Optional[str] = None,
    ) -> PrivacyRequest:
        """Creates a single privacy request.

        Args:
            privacy_request_data: Data for creating the privacy request
            authenticated: If True, bypasses identity verification
            privacy_preferences: List of privacy preferences (for consent requests only)
            user_id: Optional ID of the user submitting the request

        Returns:
            Created PrivacyRequest

        Raises:
            PrivacyRequestError: For any creation failures
        """

        if not any(privacy_request_data.identity.model_dump(mode="json").values()):
            raise PrivacyRequestError(
                "You must provide at least one identity to process",
                privacy_request_data.model_dump(mode="json"),
            )

        if privacy_request_data.property_id:
            valid_property = Property.get_by(
                self.db, field="id", value=privacy_request_data.property_id
            )
            if not valid_property:
                raise PrivacyRequestError(
                    "Property ID must be valid to process",
                    privacy_request_data.model_dump(mode="json"),
                )

        # Validate location is provided for required location fields
        self._validate_required_location_fields(privacy_request_data)

        policy = Policy.get_by(
            db=self.db,
            field="key",
            value=privacy_request_data.policy_key,
        )
        if policy is None:
            raise PrivacyRequestError(
                f"Policy with key {privacy_request_data.policy_key} does not exist",
                privacy_request_data.model_dump(mode="json"),
            )

        kwargs = build_required_privacy_request_kwargs(
            privacy_request_data.requested_at,
            policy.id,
            self.config_proxy.execution.subject_identity_verification_required,
            authenticated,
        )

        RESUBMIT_FIELDS = [
            "id",
            "reviewed_at",
            "reviewed_by",
            "identity_verified_at",
            "custom_privacy_request_fields_approved_at",
            "custom_privacy_request_fields_approved_by",
        ]

        if isinstance(privacy_request_data, PrivacyRequestResubmit):
            for field in RESUBMIT_FIELDS:
                attr = getattr(privacy_request_data, field)
                if attr is not None:
                    kwargs[field] = attr

        OPTIONAL_FIELDS = [
            "external_id",
            "started_processing_at",
            "finished_processing_at",
            "consent_preferences",
            "property_id",
            "source",
            "location",
        ]

        for field in OPTIONAL_FIELDS:
            attr = getattr(privacy_request_data, field)
            if attr is not None:
                if field == "consent_preferences":
                    attr = [consent.model_dump(mode="json") for consent in attr]
                kwargs[field] = attr

        if (
            getattr(privacy_request_data, "source")
            == PrivacyRequestSource.request_manager
        ):
            kwargs["submitted_by"] = submitted_by

        try:
            privacy_request = PrivacyRequest.create(db=self.db, data=kwargs)
            privacy_request.persist_identity(
                db=self.db, identity=privacy_request_data.identity
            )
            _create_or_update_custom_fields(
                self.db,
                privacy_request,
                privacy_request_data.consent_request_id,
                privacy_request_data.custom_privacy_request_fields,
            )

            if privacy_preferences:
                for privacy_preference in privacy_preferences:
                    privacy_preference.privacy_request_id = privacy_request.id
                    privacy_preference.save(db=self.db)

            cache_data(
                privacy_request,
                privacy_request_data.identity,
                privacy_request_data.encryption_key,
                None,
                privacy_request_data.custom_privacy_request_fields,
            )

            if masking_secrets := policy.generate_masking_secrets():
                logger.info(
                    "Caching masking secrets for privacy request {}", privacy_request.id
                )
                privacy_request.persist_masking_secrets(masking_secrets)

            check_and_dispatch_error_notifications(db=self.db)

            _handle_notifications_and_processing(
                self.db,
                self.config_proxy,
                privacy_request,
                privacy_request_data,
                policy,
                authenticated,
                self.messaging_service,
            )

            if not isinstance(privacy_request_data, PrivacyRequestResubmit):
                check_for_duplicates(db=self.db, privacy_request=privacy_request)

            return privacy_request

        except RedisNotConfigured as exc:
            logger.error(f"{exc.__class__.__name__}: {str(exc)}")
            raise exc
        except MessageDispatchException as exc:
            kwargs["privacy_request_id"] = privacy_request.id
            raise PrivacyRequestError(
                "Verification message could not be sent.", kwargs
            ) from exc
        except Exception as exc:
            logger.error(f"{exc.__class__.__name__}: {str(exc)}")
            raise PrivacyRequestError("This record could not be added", kwargs) from exc

    def create_bulk_privacy_requests(
        self,
        data: List[PrivacyRequestCreate],
        *,
        authenticated: bool = False,
        privacy_preferences: Optional[List[PrivacyPreferenceHistory]] = None,
        user_id: Optional[str] = None,
    ) -> BulkPostPrivacyRequests:
        """Creates multiple privacy requests.

        If authenticated is True the identity verification step is bypassed.
        """

        privacy_preferences = privacy_preferences or []
        created = []
        failed = []

        logger.info("Starting creation for {} privacy requests", len(data))

        for privacy_request_data in data:
            try:
                privacy_request = self.create_privacy_request(
                    privacy_request_data,
                    authenticated=authenticated,
                    privacy_preferences=privacy_preferences,
                    submitted_by=user_id,
                )
                created.append(privacy_request)
            except PrivacyRequestError as exc:
                failed.append(BulkUpdateFailed(message=exc.message, data=exc.data))

        return BulkPostPrivacyRequests(
            succeeded=created,
            failed=failed,
        )

    def resubmit_privacy_request(
        self, privacy_request_id: str
    ) -> Optional[PrivacyRequest]:
        """
        Creates a new privacy request based on an existing one and deletes the original.

        Args:
            privacy_request_id: The ID of the existing privacy request to resubmit

        Returns:
            PrivacyRequest: The newly created privacy request or None if the request is not found.
        """
        existing_privacy_request: Optional[PrivacyRequest] = self.get_privacy_request(
            privacy_request_id
        )

        if not existing_privacy_request:
            return None

        if existing_privacy_request.status == PrivacyRequestStatus.complete:
            raise FidesopsException(
                f"Cannot resubmit a {existing_privacy_request.status} privacy request"
            )

        # Copy all needed data first
        create_data = PrivacyRequestResubmit(
            id=privacy_request_id,
            external_id=existing_privacy_request.external_id,
            started_processing_at=existing_privacy_request.started_processing_at,
            finished_processing_at=existing_privacy_request.finished_processing_at,
            requested_at=existing_privacy_request.requested_at,
            identity=existing_privacy_request.get_persisted_identity(),
            custom_privacy_request_fields=existing_privacy_request.get_persisted_custom_privacy_request_fields(),
            policy_key=existing_privacy_request.policy.key,
            encryption_key=existing_privacy_request.get_cached_encryption_key(),
            property_id=existing_privacy_request.property_id,
            consent_preferences=existing_privacy_request.consent_preferences,
            source=existing_privacy_request.source,
            reviewed_at=existing_privacy_request.reviewed_at,
            reviewed_by=existing_privacy_request.reviewed_by,
            identity_verified_at=existing_privacy_request.identity_verified_at,
            custom_privacy_request_fields_approved_at=existing_privacy_request.custom_privacy_request_fields_approved_at,
            custom_privacy_request_fields_approved_by=existing_privacy_request.custom_privacy_request_fields_approved_by,
        )
        submitted_by = existing_privacy_request.submitted_by
        reviewed_by = existing_privacy_request.reviewed_by

        # Delete old request and associated data first
        self.db.query(AuditLog).filter(
            AuditLog.privacy_request_id == privacy_request_id
        ).delete()
        self.db.query(ExecutionLog).filter(
            ExecutionLog.privacy_request_id == privacy_request_id
        ).delete()
        self.db.query(RequestTask).filter(
            RequestTask.privacy_request_id == privacy_request_id
        ).delete()
        self.db.delete(existing_privacy_request)
        existing_privacy_request.clear_cached_values()

        logger.info(f"Resubmitting privacy request {privacy_request_id}")

        # Create new request
        privacy_request = self.create_privacy_request(
            create_data, authenticated=True, submitted_by=submitted_by
        )

        if _manual_approval_required(self.config_proxy, privacy_request):
            has_webhooks = self.db.query(PreApprovalWebhook).count() > 0
            # If there are pre-approval webhooks, then we do nothing since the
            # create_privacy_request method will have already triggered them,
            # and they will be responsible of approving the request.

            # If there are no pre-approval webhooks, approve the request immediately
            # since it was originally approved by a person and we don't want them to
            # have to manually re-approve
            if not has_webhooks:
                self.approve_privacy_requests(
                    [privacy_request_id],
                    reviewed_by=reviewed_by,
                    suppress_notification=True,
                )

        return privacy_request

    def approve_privacy_requests(
        self,
        request_ids: List[str],
        *,
        webhook_id: Optional[str] = None,
        reviewed_by: Optional[str] = None,
        suppress_notification: Optional[bool] = False,
    ) -> BulkReviewResponse:
        """Approve privacy requests."""
        if len(request_ids) > BULK_PRIVACY_REQUEST_BATCH_SIZE:
            raise ValueError(
                f"Bulk operations are limited to {BULK_PRIVACY_REQUEST_BATCH_SIZE} requests. "
                f"Received {len(request_ids)}. Batch requests at the caller level."
            )

        succeeded: List[PrivacyRequest] = []
        failed: List[BulkUpdateFailed] = []

        # Fetch all privacy requests in one query to avoid N+1
        privacy_requests_dict = self._fetch_privacy_requests_for_bulk_operation(
            request_ids
        )

        for request_id in request_ids:
            privacy_request = privacy_requests_dict.get(request_id)
            try:
                privacy_request = self._validate_privacy_request_for_bulk_operation(
                    privacy_request, request_id
                )
            except PrivacyRequestError as exc:
                failed.append(BulkUpdateFailed(message=exc.message, data=exc.data))
                continue

            if privacy_request.status not in [
                PrivacyRequestStatus.pending,
                PrivacyRequestStatus.duplicate,
            ]:
                failed.append(
                    BulkUpdateFailed(
                        message="Cannot transition status",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )
                continue

            try:
                now = datetime.utcnow()
                privacy_request.status = PrivacyRequestStatus.approved
                privacy_request.reviewed_at = now
                privacy_request.reviewed_by = reviewed_by

                if privacy_request.custom_fields:  # type: ignore[attr-defined]
                    privacy_request.custom_privacy_request_fields_approved_at = now
                    privacy_request.custom_privacy_request_fields_approved_by = (
                        reviewed_by
                    )

                privacy_request.save(db=self.db)

                AuditLog.create(
                    db=self.db,
                    data={
                        "privacy_request_id": privacy_request.id,
                        "action": AuditLogAction.approved,
                        "user_id": reviewed_by,
                        "webhook_id": webhook_id,  # the last webhook reply received is what approves the entire request
                    },
                )

                if not suppress_notification:
                    self.messaging_service.send_request_approved(privacy_request)
                queue_privacy_request(privacy_request.id)

                succeeded.append(privacy_request)
            except Exception as exc:
                logger.exception(exc)
                failed.append(
                    BulkUpdateFailed(
                        message="Privacy request could not be updated",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )

        return BulkReviewResponse(succeeded=succeeded, failed=failed)

    def deny_privacy_requests(
        self,
        request_ids: List[str],
        deny_reason: Optional[str],
        *,
        user_id: Optional[str] = None,
    ) -> BulkReviewResponse:
        """Deny privacy requests."""
        if len(request_ids) > BULK_PRIVACY_REQUEST_BATCH_SIZE:
            raise ValueError(
                f"Bulk operations are limited to {BULK_PRIVACY_REQUEST_BATCH_SIZE} requests. "
                f"Received {len(request_ids)}. Batch requests at the caller level."
            )

        succeeded: List[PrivacyRequest] = []
        failed: List[BulkUpdateFailed] = []

        # Fetch all privacy requests in one query to avoid N+1
        privacy_requests_dict = self._fetch_privacy_requests_for_bulk_operation(
            request_ids
        )

        for request_id in request_ids:
            privacy_request = privacy_requests_dict.get(request_id)
            try:
                privacy_request = self._validate_privacy_request_for_bulk_operation(
                    privacy_request, request_id
                )
            except PrivacyRequestError as exc:
                failed.append(BulkUpdateFailed(message=exc.message, data=exc.data))
                continue

            if privacy_request.status not in [
                PrivacyRequestStatus.pending,
                PrivacyRequestStatus.duplicate,
            ]:
                failed.append(
                    BulkUpdateFailed(
                        message="Cannot transition status",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )
                continue

            try:
                privacy_request.status = PrivacyRequestStatus.denied
                privacy_request.reviewed_at = datetime.utcnow()
                privacy_request.reviewed_by = user_id
                privacy_request.save(db=self.db)

                AuditLog.create(
                    db=self.db,
                    data={
                        "user_id": user_id,
                        "privacy_request_id": privacy_request.id,
                        "action": AuditLogAction.denied,
                        "message": deny_reason,
                    },
                )

                self.messaging_service.send_request_denied(privacy_request, deny_reason)

                succeeded.append(privacy_request)
            except Exception:
                failed.append(
                    BulkUpdateFailed(
                        message="Privacy request could not be updated",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )

        return BulkReviewResponse(succeeded=succeeded, failed=failed)

    def cancel_privacy_requests(
        self,
        request_ids: List[str],
        cancel_reason: Optional[str],
        *,
        user_id: Optional[str] = None,
    ) -> BulkReviewResponse:
        """Cancel a list of privacy requests and/or report failure"""
        if len(request_ids) > BULK_PRIVACY_REQUEST_BATCH_SIZE:
            raise ValueError(
                f"Bulk operations are limited to {BULK_PRIVACY_REQUEST_BATCH_SIZE} requests. "
                f"Received {len(request_ids)}. Batch requests at the caller level."
            )

        succeeded: List[PrivacyRequest] = []
        failed: List[BulkUpdateFailed] = []

        # Terminal states that cannot be canceled
        terminal_states = [
            PrivacyRequestStatus.complete,
            PrivacyRequestStatus.denied,
            PrivacyRequestStatus.canceled,
            PrivacyRequestStatus.error,
        ]

        # Fetch all privacy requests in one query to avoid N+1
        privacy_requests_dict = self._fetch_privacy_requests_for_bulk_operation(
            request_ids
        )

        for request_id in request_ids:
            privacy_request = privacy_requests_dict.get(request_id)
            try:
                privacy_request = self._validate_privacy_request_for_bulk_operation(
                    privacy_request, request_id
                )
            except PrivacyRequestError as exc:
                failed.append(BulkUpdateFailed(message=exc.message, data=exc.data))
                continue

            if privacy_request.status in terminal_states:
                failed.append(
                    BulkUpdateFailed(
                        message=f"Cannot cancel privacy request in {privacy_request.status.value} status",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )
                continue

            try:
                privacy_request.cancel_processing(
                    db=self.db, cancel_reason=cancel_reason
                )

                succeeded.append(privacy_request)
            except Exception:
                failed.append(
                    BulkUpdateFailed(
                        message="Privacy request could not be canceled",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )

        return BulkReviewResponse(succeeded=succeeded, failed=failed)

    def finalize_privacy_requests(
        self,
        request_ids: List[str],
        *,
        user_id: Optional[str] = None,
    ) -> BulkReviewResponse:
        """Finalize privacy requests."""
        if len(request_ids) > BULK_PRIVACY_REQUEST_BATCH_SIZE:
            raise ValueError(
                f"Bulk operations are limited to {BULK_PRIVACY_REQUEST_BATCH_SIZE} requests. "
                f"Received {len(request_ids)}. Batch requests at the caller level."
            )

        succeeded: List[PrivacyRequest] = []
        failed: List[BulkUpdateFailed] = []

        # Fetch all privacy requests in one query to avoid N+1
        privacy_requests_dict = self._fetch_privacy_requests_for_bulk_operation(
            request_ids
        )

        for request_id in request_ids:
            privacy_request = privacy_requests_dict.get(request_id)
            try:
                privacy_request = self._validate_privacy_request_for_bulk_operation(
                    privacy_request, request_id
                )
            except PrivacyRequestError as exc:
                failed.append(BulkUpdateFailed(message=exc.message, data=exc.data))
                continue

            if (
                privacy_request.status
                != PrivacyRequestStatus.requires_manual_finalization
            ):
                failed.append(
                    BulkUpdateFailed(
                        message=f"Cannot manually finalize privacy request: status is {privacy_request.status}, not requires_manual_finalization",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )
                continue

            try:
                # Set finalized_by and finalized_at here, so the request runner service knows not to
                # put the request back into the requires_finalization state.
                privacy_request.finalized_at = datetime.utcnow()
                privacy_request.finalized_by = user_id
                privacy_request.save(db=self.db)

                # Queue the privacy request for finalization
                queue_privacy_request(
                    privacy_request_id=privacy_request.id,
                    from_step=CurrentStep.finalization.value,
                )

                succeeded.append(privacy_request)
            except Exception as exc:
                logger.exception(exc)
                failed.append(
                    BulkUpdateFailed(
                        message="Privacy request could not be finalized",
                        data=PrivacyRequestResponse.model_validate(
                            privacy_request
                        ).model_dump(mode="json"),
                    )
                )

        return BulkReviewResponse(succeeded=succeeded, failed=failed)


@log_context(capture_args={"privacy_request_id": LoggerContextKeys.privacy_request_id})
def queue_privacy_request(
    privacy_request_id: str,
    from_webhook_id: Optional[str] = None,
    from_step: Optional[str] = None,
) -> str:
    logger.info("Queueing privacy request from step {}", from_step)

    from fides.api.service.privacy_request.request_runner_service import (
        run_privacy_request,
    )

    task = run_privacy_request.apply_async(
        queue=DSR_QUEUE_NAME,
        kwargs={
            "privacy_request_id": privacy_request_id,
            "from_webhook_id": from_webhook_id,
            "from_step": from_step,
        },
    )
    cache_task_tracking_key(privacy_request_id, task.task_id)

    return task.task_id


def _create_or_update_custom_fields(
    db: Session,
    privacy_request: PrivacyRequest,
    consent_request_id: Optional[str],
    custom_privacy_request_fields: Optional[Dict[str, Any]],
) -> None:
    """
    Updates existing custom privacy request fields in the database with a privacy request ID.
    Creates new custom privacy request fields if there aren't any available.

    The presence or absence of custom fields is based on whether or not the creation of this
    current privacy request was triggered by a consent request.
    """
    consent_request = ConsentRequest.get_by_key_or_id(
        db=db, data={"id": consent_request_id}
    )
    if consent_request and consent_request.custom_fields:
        for custom_field in consent_request.custom_fields:  # type: ignore[attr-defined]
            custom_field.privacy_request_id = privacy_request.id
            custom_field.save(db=db)
    elif custom_privacy_request_fields:
        privacy_request.persist_custom_privacy_request_fields(
            db=db,
            custom_privacy_request_fields=custom_privacy_request_fields,
        )


def _handle_notifications_and_processing(
    db: Session,
    config_proxy: ConfigProxy,
    privacy_request: PrivacyRequest,
    privacy_request_data: PrivacyRequestCreate,
    policy: Policy,
    authenticated: bool,
    messaging_service: MessagingService,
) -> None:
    """Handle notifications and request processing after creation"""
    if not authenticated and message_send_enabled(
        db,
        privacy_request.property_id,
        MessagingActionType.SUBJECT_IDENTITY_VERIFICATION,
        config_proxy.execution.subject_identity_verification_required,
    ):
        messaging_service.send_verification_code(
            privacy_request,
            privacy_request_data.identity,
            privacy_request.property_id,
        )
        return

    if not authenticated and message_send_enabled(
        db,
        privacy_request.property_id,
        MessagingActionType.PRIVACY_REQUEST_RECEIPT,
        config_proxy.notifications.send_request_receipt_notification,
    ):
        send_privacy_request_receipt_message_to_user(
            policy,
            privacy_request_data.identity,
            config_proxy.notifications.notification_service_type,
            privacy_request.property_id,
        )

    handle_approval(db, config_proxy, privacy_request)


def _trigger_pre_approval_webhooks(
    db: Session, privacy_request: PrivacyRequest
) -> None:
    """
    Shared method to trigger all configured pre-approval webhooks for a given privacy request.
    """
    pre_approval_webhooks = db.query(PreApprovalWebhook).all()
    for webhook in pre_approval_webhooks:
        privacy_request.trigger_pre_approval_webhook(
            webhook=webhook,
            policy_action=privacy_request.policy.get_action_type(),
        )


def _requeue_privacy_request(
    db: Session,
    privacy_request: PrivacyRequest,
) -> PrivacyRequestResponse:
    """If failed_step is provided, restart the DSR within that step. Otherwise,
    restart the privacy request from the beginning."""

    if privacy_request.status not in [
        PrivacyRequestStatus.approved,
        PrivacyRequestStatus.in_processing,
    ]:
        raise PrivacyRequestError(
            f"Cannot re-queue privacy request {privacy_request.id} with status {privacy_request.status.value}"
        )

    # Both DSR 2.0 and 3.0 cache checkpoint details
    checkpoint_details: Optional[CheckpointActionRequired] = (
        privacy_request.get_failed_checkpoint_details()
    )
    resume_step = checkpoint_details.step if checkpoint_details else None

    # DSR 3.0 additionally stores Request Tasks in the application db that can be used to infer
    # a resume checkpoint in the event the cache has expired.
    if not resume_step and privacy_request.request_tasks.count():
        if privacy_request.consent_tasks.count():
            resume_step = CurrentStep.consent
        elif privacy_request.erasure_tasks.count():
            # Checking if access terminator task was completed, because erasure tasks are created
            # at the same time as the access tasks
            terminator_access_task = privacy_request.get_terminate_task_by_action(
                ActionType.access
            )
            resume_step = (
                CurrentStep.erasure
                if terminator_access_task.status == ExecutionLogStatus.complete
                else CurrentStep.access
            )
        elif privacy_request.access_tasks.count():
            resume_step = CurrentStep.access

    logger.debug(
        "Re-queuing privacy request {} from step {}",
        privacy_request.id,
        resume_step.value if resume_step else None,
    )

    return _process_privacy_request_restart(
        privacy_request,
        resume_step,
        db,
    )


def _process_privacy_request_restart(
    privacy_request: PrivacyRequest,
    failed_step: Optional[CurrentStep],
    db: Session,
) -> PrivacyRequestResponse:
    """If failed_step is provided, restart the DSR within that step. Otherwise,
    restart the privacy request from the beginning."""
    if failed_step:
        logger.info(
            "Restarting failed privacy request '{}' from '{}'",
            privacy_request.id,
            failed_step,
        )
    else:
        logger.info(
            "Restarting failed privacy request '{}' from the beginning",
            privacy_request.id,
        )

    privacy_request.status = PrivacyRequestStatus.in_processing
    privacy_request.save(db=db)
    queue_privacy_request(
        privacy_request_id=privacy_request.id,
        from_step=failed_step.value if failed_step else None,
    )

    return privacy_request  # type: ignore[return-value]


def _manual_approval_required(
    config_proxy: ConfigProxy,
    privacy_request: PrivacyRequest,
) -> bool:
    """Determines if a privacy request requires manual approval."""
    return (
        config_proxy.execution.require_manual_request_approval
        and privacy_request.policy.get_action_type() != ActionType.consent.value
    )


def handle_approval(
    db: Session, config_proxy: ConfigProxy, privacy_request: PrivacyRequest
) -> None:
    """Evaluate manual approval and handle processing or pre-approval webhooks."""
    check_for_duplicates(db=db, privacy_request=privacy_request)
    if privacy_request.status == PrivacyRequestStatus.duplicate:
        return
    if _manual_approval_required(config_proxy, privacy_request):
        _trigger_pre_approval_webhooks(db, privacy_request)
    else:
        AuditLog.create(
            db=db,
            data={
                "user_id": "system",
                "privacy_request_id": privacy_request.id,
                "action": AuditLogAction.approved,
                "message": "",
            },
        )
        queue_privacy_request(privacy_request.id)



from typing import Any, Dict


def mask_sensitive_fields(
    connection_secrets: Dict[str, Any], secret_schema: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Mask sensitive fields in the given secrets based on the provided schema.
    This function traverses the given secrets dictionary and uses the provided schema to
    identify fields that have been marked as sensitive. The function replaces the sensitive
    field values with a mask string ('********').
    Args:
        connection_secrets (Dict[str, Any]): The secrets to be masked.
        secret_schema (Dict[str, Any]): The schema defining which fields are sensitive.
    Returns:
        Dict[str, Any]: The secrets dictionary with sensitive fields masked.
    """
    if connection_secrets is None:
        return connection_secrets

    secret_schema_properties = secret_schema["properties"]
    new_connection_secrets = {}

    for key, value in connection_secrets.items():
        if key in secret_schema_properties:
            if secret_schema_properties.get(key, {}).get("sensitive", False):
                new_connection_secrets[key] = "**********"
            else:
                new_connection_secrets[key] = value

    return new_connection_secrets



from datetime import datetime, timedelta, timezone
from typing import Optional
from uuid import UUID

from loguru import logger
from sqlalchemy.orm import Session, defer

from fides.api.models.privacy_request.duplicate_group import (
    DuplicateGroup,
    generate_rule_version,
)
from fides.api.models.privacy_request.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.task.conditional_dependencies.schemas import (
    Condition,
    ConditionGroup,
    ConditionLeaf,
    GroupOperator,
    Operator,
)
from fides.api.task.conditional_dependencies.sql_translator import (
    SQLConditionTranslator,
)
from fides.config.config_proxy import ConfigProxy
from fides.config.duplicate_detection_settings import DuplicateDetectionSettings

ACTIONED_REQUEST_STATUSES = [
    PrivacyRequestStatus.approved,
    PrivacyRequestStatus.in_processing,
    PrivacyRequestStatus.requires_manual_finalization,
    PrivacyRequestStatus.requires_input,
    PrivacyRequestStatus.paused,
    PrivacyRequestStatus.awaiting_email_send,
    PrivacyRequestStatus.error,
]


class DuplicateDetectionService:
    def __init__(self, db: Session):
        self.db = db
        self._config = ConfigProxy(db).privacy_request_duplicate_detection

    def is_enabled(self) -> bool:
        return self._config.enabled

    def _create_identity_conditions(
        self, current_request: PrivacyRequest
    ) -> list[Condition]:
        """Creates conditions for matching identity fields.

        For identity field matching using the EAV pattern in ProvidedIdentity, we need to match both field_name
        and hashed_value. This function creates the required nested conditions for each identity field.
        Also adds a condition for the policy_id to ensure that we are only matching requests for the same policy.
        """
        conditions: list[Condition] = []
        current_identities: dict[str, str] = {
            pi.field_name: pi.hashed_value
            for pi in current_request.provided_identities  # type: ignore [attr-defined]
            if pi.field_name in self._config.match_identity_fields
        }
        if len(current_identities) != len(self._config.match_identity_fields):
            missing_fields = [
                field
                for field in self._config.match_identity_fields
                if field not in current_identities.keys()
            ]
            logger.debug(
                f"Some identity fields were not found in the current request: {missing_fields}"
            )
            return []

        for field_name, hashed_value in current_identities.items():
            identity_condition = ConditionGroup(
                logical_operator=GroupOperator.and_,
                conditions=[
                    ConditionLeaf(
                        field_address="privacyrequest.provided_identities.field_name",
                        operator=Operator.eq,
                        value=field_name,
                    ),
                    ConditionLeaf(
                        field_address="privacyrequest.provided_identities.hashed_value",
                        operator=Operator.eq,
                        value=hashed_value,
                    ),
                ],
            )
            conditions.append(identity_condition)
        policy_condition = ConditionLeaf(
            field_address="privacyrequest.policy_id",
            operator=Operator.eq,
            value=current_request.policy_id,
        )
        conditions.append(policy_condition)
        return conditions

    def _create_time_window_condition(self, time_window_days: int) -> Condition:
        """Creates a condition for matching requests within a time window."""
        cutoff_date = datetime.now(timezone.utc) - timedelta(days=time_window_days)
        condition = ConditionLeaf(
            field_address="privacyrequest.created_at",
            operator=Operator.gte,
            value=cutoff_date.isoformat(),
        )
        return condition

    def create_duplicate_detection_conditions(
        self,
        current_request: PrivacyRequest,
    ) -> Optional[ConditionGroup]:
        """
        Create conditions for duplicate detection based on configuration.

        Args:
            current_request: The current privacy request to find duplicates for
            config: Duplicate detection configuration settings

        Returns:
            A ConditionGroup with AND operator, or None if no conditions can be created
        """
        if len(self._config.match_identity_fields) == 0:
            return None

        identity_conditions = self._create_identity_conditions(current_request)
        if not identity_conditions:
            return None  # Only proceed if we have identity conditions

        time_window_condition = self._create_time_window_condition(
            self._config.time_window_days
        )

        # Combine all conditions with AND operator
        all_conditions: list[Condition] = [*identity_conditions, time_window_condition]
        return ConditionGroup(
            logical_operator=GroupOperator.and_, conditions=all_conditions
        )

    def find_duplicate_privacy_requests(
        self,
        current_request: PrivacyRequest,
    ) -> list[PrivacyRequest]:
        """
        Find potential duplicate privacy requests based on duplicate detection configuration.

        Uses the SQLConditionTranslator to build queries from conditions, which handles
        the ProvidedIdentity relationship using SQLAlchemy's .any() method.

        Args:
            current_request: The privacy request to check for duplicates
            config: Duplicate detection configuration settings

        Returns:
            List of PrivacyRequest objects that match the duplicate criteria,
            does not include the current request
        """
        condition = self.create_duplicate_detection_conditions(current_request)

        if condition is None:
            return []

        translator = SQLConditionTranslator(self.db)
        query = translator.generate_query_from_condition(condition)

        query = query.filter(PrivacyRequest.id != current_request.id).filter(
            PrivacyRequest.deleted_at.is_(None)
        )
        # Apply defer options to prevent loading large columns when checking for duplicates
        # pylint: disable=protected-access
        query = query.options(
            defer(PrivacyRequest._filtered_final_upload),
            defer(PrivacyRequest.access_result_urls),
        )
        return query.all()

    def generate_dedup_key(self, request: PrivacyRequest) -> str:
        """
        Generate a dedup key for a request based on the duplicate detection settings.
        """
        current_identities: dict[str, str] = {
            pi.field_name: pi.hashed_value
            for pi in request.provided_identities  # type: ignore [attr-defined]
            if pi.field_name in self._config.match_identity_fields
        }
        if len(current_identities) != len(self._config.match_identity_fields):
            raise ValueError(
                "This request does not contain the required identity fields for duplicate detection."
            )
        return "|".join(
            [
                current_identities[field]
                for field in sorted(self._config.match_identity_fields)
            ]
        )

    def update_duplicate_group_ids(
        self,
        request: PrivacyRequest,
        duplicates: list[PrivacyRequest],
        duplicate_group_id: UUID,
    ) -> None:
        """
        Update the duplicate request group ids for a request and its duplicates.
        Args:
            request: The privacy request to update
            duplicates: The list of duplicate requests to update
            duplicate_group_id: The duplicate request group id to update
        """
        update_all = [request] + duplicates
        try:
            for privacy_request in update_all:
                privacy_request.duplicate_request_group_id = duplicate_group_id  # type: ignore [assignment]
        except Exception as e:
            logger.error(f"Failed to update duplicate request group ids: {e}")
            raise e

    def mark_as_duplicate(self, request: PrivacyRequest, message: str) -> None:
        """
        Mark a request as a duplicate.
        """
        request.update(self.db, data={"status": PrivacyRequestStatus.duplicate})
        logger.debug(message)
        request.add_skipped_execution_log(
            self.db,
            connection_key=None,
            dataset_name="Duplicate Request Detection",
            collection_name=None,
            message=message,
            action_type=(request.policy.get_action_type() if request.policy else None)
            or ActionType.access,
        )

    def add_success_execution_log(self, request: PrivacyRequest, message: str) -> None:
        request.add_success_execution_log(
            db=self.db,
            connection_key=None,
            dataset_name="Duplicate Request Detection",
            collection_name=None,
            message=message,
            action_type=(
                request.policy.get_action_type()  # type: ignore [arg-type]
                if request.policy
                else ActionType.access
            ),
        )

    def verified_identity_cases(
        self, request: PrivacyRequest, duplicates: list[PrivacyRequest]
    ) -> bool:
        """
        Apply verified identity rules to determine if a request is a duplicate request.
        - If this request does not have a verified identity, it may be a duplicate if another request in the group is verified.
        - If this is the first request to be verified, it is not a duplicate request
        - If other requests identities were verified before this request, it is a duplicate request
        Args:
            request: The privacy request to check
            duplicates: The list of duplicate requests

        Returns:
            True if the request is a duplicate request, False otherwise
        """
        verified_in_group = [
            duplicate for duplicate in duplicates if duplicate.identity_verified_at
        ]

        # The request identity is not verified.
        if not request.identity_verified_at:
            if len(verified_in_group) > 0:
                message = f"Request {request.id} is a duplicate: it is duplicating request(s) {[duplicate.id for duplicate in verified_in_group]}."
                self.mark_as_duplicate(request, message)
                return True

            canonical_request = min(duplicates, key=lambda x: x.created_at)  # type: ignore [arg-type, return-value]
            canonical_request_created_at = canonical_request.created_at or datetime.now(
                timezone.utc
            )
            request_created_at = request.created_at or datetime.now(timezone.utc)
            if request_created_at < canonical_request_created_at:
                message = f"Request {request.id} is not a duplicate: it is the first request to be created in the group."
                logger.debug(message)
                self.add_success_execution_log(request, message)
                return False

            message = f"Request {request.id} is a duplicate: it is duplicating request(s) ['{canonical_request.id}']."
            self.mark_as_duplicate(request, message)
            return True

        # The request identity is verified.
        if not verified_in_group:
            message = f"Request {request.id} is not a duplicate: it is the first request to be verified in the group."
            logger.debug(message)
            for duplicate in duplicates:
                dup_message = f"Request {duplicate.id} is a duplicate: it is duplicating request(s) ['{request.id}']."
                self.mark_as_duplicate(duplicate, dup_message)
            self.add_success_execution_log(request, message)
            return False

        # If this request is the first with a verified identity, it is not a duplicate.
        canonical_request = min(verified_in_group, key=lambda x: x.identity_verified_at)  # type: ignore [arg-type, return-value]
        canonical_request_verified_at = (
            canonical_request.identity_verified_at or datetime.now(timezone.utc)
        )
        request_verified_at = request.identity_verified_at or datetime.now(timezone.utc)
        if request_verified_at < canonical_request_verified_at:
            message = f"Request {request.id} is not a duplicate: it is the first request to be verified in the group."
            logger.debug(message)
            self.add_success_execution_log(request, message)
            for duplicate in duplicates:
                dup_message = f"Request {duplicate.id} is a duplicate: it is duplicating request(s) ['{request.id}']."
                self.mark_as_duplicate(duplicate, dup_message)
            return False
        message = f"Request {request.id} is a duplicate: it is duplicating request(s) ['{canonical_request.id}']."
        self.mark_as_duplicate(request, message)
        return True

    # pylint: disable=too-many-return-statements
    def is_duplicate_request(self, request: PrivacyRequest) -> bool:
        """
        Determine if a request is a duplicate request and assigns a duplicate request group id.

        The hierarchy is:
        1. Actioned requests: if this request duplicates an actioned request, it is a duplicate.
        2. Verified identity requests:
          a. if this request has a verified identity:
            - If none of the duplicates have a verified identity, it is not a duplicate.
            - If duplicates have verified identities, but this request is the first with a verified identity, it is not a duplicate.
          b. if this request does not have a verified identity:
            - If no duplicates have a verified identity, and this was the first created request, it is not a duplicate.
        3. First created request: if this is the first created request in the group, it is not a duplicate.
        4. If no canonical requests are found (meaning all requests are marked as duplicates), this request is not a duplicate.
            - Could occur if configuration changes and previous requests were already marked as duplicates.

        Args:
            request: The privacy request to check
            config: Duplicate detection configuration settings
        Returns:
            True if the request is a duplicate request, False otherwise
        """
        if request.policy.get_action_type() == ActionType.consent:
            message = f"Consent request {request.id} is not a duplicate."
            logger.info(message)
            self.add_success_execution_log(request, message)
            return False
        duplicates = self.find_duplicate_privacy_requests(request)
        rule_version = generate_rule_version(
            DuplicateDetectionSettings(
                enabled=self._config.enabled,
                time_window_days=self._config.time_window_days,
                match_identity_fields=self._config.match_identity_fields,
            )
        )
        try:
            dedup_key = self.generate_dedup_key(request)
        except ValueError as e:
            message = f"Request {request.id} is not a duplicate: {e}"
            logger.warning(message)
            self.add_success_execution_log(request, message)
            return False

        _, duplicate_group = DuplicateGroup.get_or_create(
            db=self.db, data={"rule_version": rule_version, "dedup_key": dedup_key}
        )
        if duplicate_group is None:
            message = f"Failed to create duplicate group for request {request.id} with dedup key {dedup_key}"
            logger.error(message)
            request.add_error_execution_log(
                db=self.db,
                connection_key=None,
                dataset_name="Duplicate Request Detection",
                collection_name=None,
                message=message,
                action_type=(
                    request.policy.get_action_type()  # type: ignore [arg-type]
                    if request.policy
                    else ActionType.access
                ),
            )
            return False

        self.update_duplicate_group_ids(request, duplicates, duplicate_group.id)  # type: ignore [arg-type]

        if request.status in ACTIONED_REQUEST_STATUSES:
            message = (
                f"Request {request.id} is not a duplicate: it is an actioned request."
            )
            logger.debug(message)
            self.add_success_execution_log(request, message)
            return False

        # if this is the only request in the group, it is not a duplicate
        if len(duplicates) == 0:
            message = f"Request {request.id} is not a duplicate."
            logger.debug(message)
            self.add_success_execution_log(request, message)
            return False

        if request.status == PrivacyRequestStatus.duplicate:
            return True

        # only compare to non-duplicate/complete requests for the following cases
        canonical_requests = [
            duplicate
            for duplicate in duplicates
            if duplicate.status
            not in [
                PrivacyRequestStatus.duplicate,
                PrivacyRequestStatus.complete,
                PrivacyRequestStatus.denied,
            ]
        ]
        # If no non-duplicate requests are found, this request is not a duplicate.
        if len(canonical_requests) == 0:
            message = f"Request {request.id} is not a duplicate."
            logger.debug(message)
            self.add_success_execution_log(request, message)
            return False

        # If any requests in group are actioned, this request is a duplicate.
        actioned_in_group = [
            duplicate
            for duplicate in canonical_requests
            if duplicate.status in ACTIONED_REQUEST_STATUSES
        ]
        if len(actioned_in_group) > 0:
            message = f"Request {request.id} is a duplicate: it is duplicating actioned request(s) {[duplicate.id for duplicate in actioned_in_group]}."
            self.mark_as_duplicate(request, message)
            return True
        # Check against verified identity rules.
        return self.verified_identity_cases(request, canonical_requests)


def check_for_duplicates(db: Session, privacy_request: PrivacyRequest) -> None:
    duplicate_detection_service = DuplicateDetectionService(db)
    if duplicate_detection_service.is_enabled():
        logger.info(
            "Duplicate detection is enabled. Checking if privacy request is a duplicate."
        )
        if duplicate_detection_service.is_duplicate_request(privacy_request):
            logger.info("Terminating privacy request: request is a duplicate.")



from datetime import datetime
from typing import Any, Dict, List, Optional, cast

from fideslang.models import Dataset
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import BaseModel, ConfigDict, model_validator

from fides.api.common_exceptions import NoSuchConnectionTypeSecretSchemaError
from fides.api.models.connectionconfig import AccessLevel, ConnectionType
from fides.api.schemas.api import BulkResponse, BulkUpdateFailed
from fides.api.schemas.connection_configuration import connection_secrets_schemas
from fides.api.schemas.policy import ActionType
from fides.api.schemas.saas.saas_config import SaaSConfigBase
from fides.api.util.connection_type import get_connection_type_secret_schema
from fides.api.util.masking_util import mask_sensitive_fields


class CreateConnectionConfiguration(BaseModel):
    """
    Schema for creating a ConnectionConfiguration

    Note that secrets are *NOT* allowed to be supplied here.
    """

    name: Optional[str] = None
    key: Optional[FidesKey] = None
    connection_type: ConnectionType
    access: AccessLevel
    disabled: Optional[bool] = False
    description: Optional[str] = None
    model_config = ConfigDict(
        from_attributes=True, use_enum_values=True, extra="ignore"
    )


class CreateConnectionConfigurationWithSecrets(CreateConnectionConfiguration):
    """Schema for creating a connection configuration including secrets."""

    secrets: Optional[connection_secrets_schemas] = None
    saas_connector_type: Optional[str] = None
    model_config = ConfigDict(from_attributes=True, extra="ignore")


class ConnectionConfigSecretsMixin(BaseModel):
    """
    A schema mixin to declare a connection config `secrets` attribute
    and handle masking of sensitive values based on `connection_type`
    and (optionally) a `saas_config`.
    """

    connection_type: ConnectionType
    secrets: Optional[Dict[str, Any]] = None
    saas_config: Optional[SaaSConfigBase] = None

    @model_validator(mode="after")
    def mask_sensitive_values(self) -> "ConnectionConfigSecretsMixin":
        """
        Mask sensitive values in the `secrets` attribute based on `connection_type`
        and (optionally) a `saas_config`.
        """
        if self.secrets is None:
            return self

        connection_type = (
            self.saas_config.type
            if self.connection_type == ConnectionType.saas and self.saas_config
            else self.connection_type.value  # type: ignore
        )
        try:
            secret_schema = get_connection_type_secret_schema(
                connection_type=connection_type
            )
        except NoSuchConnectionTypeSecretSchemaError as e:
            logger.error(e)
            # if there is no schema, we don't know what values to mask.
            # so all the secrets are removed.
            self.secrets = None
            return self

        self.secrets = mask_sensitive_fields(cast(dict, self.secrets), secret_schema)
        return self


class ConnectionConfigurationResponseBase(ConnectionConfigSecretsMixin):
    """
    Base schema for ConnectionConfiguration responses,
    excluding fields that might be conditionally omitted.
    The mixin base class ensures that `secrets` sensitive values are masked.
    """

    name: Optional[str] = None
    key: FidesKey
    description: Optional[str] = None
    access: AccessLevel
    created_at: datetime
    updated_at: Optional[datetime] = None
    disabled: Optional[bool] = False
    last_test_timestamp: Optional[datetime] = None
    last_test_succeeded: Optional[bool] = None
    authorized: Optional[bool] = False
    enabled_actions: Optional[List[ActionType]] = None
    model_config = ConfigDict(from_attributes=True)


class ConnectionConfigurationResponseWithSystemKey(ConnectionConfigurationResponseBase):
    """
    Describes the full returned schema for a ConnectionConfiguration.
    """

    # Using this response with models returned from an async DB session will error
    # because the system_key is lazy loaded. Just a quirk of SQLAlchemy 1.4.
    system_key: Optional[str] = None


class ConnectionConfigurationResponse(ConnectionConfigurationResponseBase):
    """
    Describes the returned schema for a ConnectionConfiguration.
    """


class BulkPutConnectionConfiguration(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of ConnectionConfiguration responses."""

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]


class BulkPatchConnectionConfigurationWithSecrets(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of ConnectionConfigurationWithSecrets responses."""

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]


class SaasConnectionTemplateResponse(BaseModel):
    connection: ConnectionConfigurationResponse
    dataset: Dataset



from typing import Annotated, List, Optional

from fastapi import Depends, HTTPException
from fastapi.encoders import jsonable_encoder
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field, ValidationError
from sqlalchemy.orm import Session
from starlette.status import HTTP_404_NOT_FOUND, HTTP_422_UNPROCESSABLE_ENTITY

from fides.api.api import deps
from fides.api.common_exceptions import (
    ConnectionNotFoundException,
    KeyOrNameAlreadyExists,
    SaaSConfigNotFoundException,
)
from fides.api.common_exceptions import ValidationError as FidesValidationError
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.manual_webhook import AccessManualWebhook
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.sql_models import Dataset as CtlDataset  # type: ignore
from fides.api.models.sql_models import System  # type: ignore
from fides.api.schemas.api import BulkUpdateFailed
from fides.api.schemas.connection_configuration import (
    ConnectionConfigSecretsSchema,
    connection_secrets_schemas,
)
from fides.api.schemas.connection_configuration.connection_config import (
    BulkPutConnectionConfiguration,
    ConnectionConfigurationResponse,
    CreateConnectionConfigurationWithSecrets,
)
from fides.api.schemas.connection_configuration.connection_secrets import (
    TestStatusMessage,
)
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.common.api.v1.urn_registry import SAAS_CONFIG
from fides.service.connection.connection_service import (
    ConnectionService,
    ConnectorTemplateNotFound,
)
from fides.service.event_audit_service import EventAuditService
from fides.service.privacy_request.privacy_request_service import queue_privacy_request

# pylint: disable=too-many-nested-blocks,too-many-branches,too-many-statements


def requeue_requires_input_requests(db: Session) -> None:
    """
    Queue privacy requests with request status "requires_input" if they are no longer blocked by
    access manual webhooks.

    For use when all access manual webhooks have been either disabled or deleted, leaving privacy requests
    lingering in a "requires_input" state.
    """
    if not AccessManualWebhook.get_enabled(db):
        for pr in PrivacyRequest.query_without_large_columns(db).filter(
            PrivacyRequest.status == PrivacyRequestStatus.requires_input
        ):
            logger.info(
                "Queuing privacy request '{} with '{}' status now that manual inputs are no longer required.",
                pr.id,
                pr.status.value,
            )
            pr.status = PrivacyRequestStatus.in_processing
            pr.save(db=db)
            queue_privacy_request(
                privacy_request_id=pr.id,
            )


def validate_secrets_error_message() -> str:
    return f"A SaaS config to validate the secrets is unavailable for this connection config, please add one via {SAAS_CONFIG}"


def validate_secrets(
    db: Session,
    request_body: connection_secrets_schemas,
    connection_config: ConnectionConfig,
) -> ConnectionConfigSecretsSchema:
    """Validate incoming connection configuration secrets."""

    event_audit_service = EventAuditService(db)
    connection_service = ConnectionService(db, event_audit_service)

    try:
        return connection_service.validate_secrets(request_body, connection_config)
    except SaaSConfigNotFoundException:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=validate_secrets_error_message(),
        )
    except FidesValidationError as e:
        # Check if the exception has the original pydantic errors attached
        raise HTTPException(status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=e.message)
    except ValidationError as e:
        errors = e.errors(include_url=False, include_input=False)
        for err in errors:
            # Additionally, manually remove the context from the error message -
            # this may contain sensitive information
            err.pop("ctx", None)
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(errors),
        )


def patch_connection_configs(
    db: Session,
    configs: Annotated[List[CreateConnectionConfigurationWithSecrets], Field(max_length=50)],  # type: ignore
    system: Optional[System] = None,
) -> BulkPutConnectionConfiguration:
    created_or_updated: List[ConnectionConfigurationResponse] = []
    failed: List[BulkUpdateFailed] = []
    logger.info("Starting bulk upsert for {} connection configuration(s)", len(configs))

    # Initialize the connection service
    event_audit_service = EventAuditService(db)
    connection_service = ConnectionService(db, event_audit_service)

    for config in configs:
        orig_data = config.model_dump(serialize_as_any=True, mode="json").copy()

        try:
            # Use the service to handle individual connection config processing
            connection_response = connection_service.create_or_update_connection_config(
                config, system
            )
            created_or_updated.append(connection_response)

        except KeyOrNameAlreadyExists as exc:
            logger.warning(
                "Create/update failed for connection config with key '{}': {}",
                config.key,
                exc,
            )
            # remove secrets information from the return for security reasons.
            orig_data.pop("secrets", None)
            orig_data.pop("saas_connector_type", None)
            failed.append(
                BulkUpdateFailed(
                    message=exc.args[0],
                    data=orig_data,
                )
            )
        except ConnectorTemplateNotFound as exc:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail=str(exc),
            )
        except ValidationError as e:
            # The "input" potentially contains sensitive info and the Pydantic-specific "url" is not helpful
            errors = e.errors(include_url=False, include_input=False)
            for err in errors:
                # Additionally, manually remove the context from the error message -
                # this may contain sensitive information
                err.pop("ctx", None)

            raise HTTPException(
                status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                detail=jsonable_encoder(errors),
            )
        except ValueError as exc:
            # Handle missing saas_connector_type and other validation errors
            raise HTTPException(
                status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                detail=str(exc),
            )
        except Exception as e:
            logger.warning(
                "Create/update failed for connection config with key '{}'.", config.key
            )
            logger.error(e)
            # remove secrets information from the return for security reasons.
            orig_data.pop("secrets", None)
            orig_data.pop("saas_connector_type", None)
            failed.append(
                BulkUpdateFailed(
                    message="This connection configuration could not be added.",
                    data=orig_data,
                )
            )

    # Check if possibly disabling a manual webhook here causes us to need to queue affected privacy requests
    requeue_requires_input_requests(db)

    logger.info(
        "Completed bulk upsert for {} connection configuration(s): {} succeeded, {} failed",
        len(configs),
        len(created_or_updated),
        len(failed),
    )
    return BulkPutConnectionConfiguration(
        succeeded=created_or_updated,
        failed=failed,
    )


def get_connection_config_or_error(
    db: Session, connection_key: FidesKey
) -> ConnectionConfig:
    """Helper to load the ConnectionConfig object or throw a 404"""
    connection_config = ConnectionConfig.get_by(db, field="key", value=connection_key)
    logger.info("Finding connection configuration with key '{}'", connection_key)
    if not connection_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connection configuration found with key '{connection_key}'.",
        )
    return connection_config


def delete_connection_config(db: Session, connection_key: FidesKey) -> None:
    """Removes the connection configuration with matching key."""
    connection_config = get_connection_config_or_error(db, connection_key)
    connection_type = connection_config.connection_type
    logger.info("Deleting connection config with key '{}'.", connection_key)

    # Initialize services for audit logging
    event_audit_service = EventAuditService(db)
    connection_service = ConnectionService(db, event_audit_service)

    # Handle SaaS-specific cleanup
    if connection_config.saas_config:
        saas_dataset_fides_key = connection_config.saas_config.get("fides_key")

        dataset_config = db.query(DatasetConfig).filter(
            DatasetConfig.connection_config_id == connection_config.id
        )
        dataset_config.delete(synchronize_session="evaluate")

        if saas_dataset_fides_key:
            logger.info("Deleting saas dataset with key '{}'.", saas_dataset_fides_key)
            saas_dataset = (
                db.query(CtlDataset)
                .filter(CtlDataset.fides_key == saas_dataset_fides_key)
                .first()
            )
            saas_dataset.delete(db)  # type: ignore[union-attr]

    # Use connection service to delete with audit logging
    connection_service.delete_connection_config(connection_key)

    # Access Manual Webhooks are cascade deleted if their ConnectionConfig is deleted,
    # so we queue any privacy requests that are no longer blocked by webhooks
    if connection_type == ConnectionType.manual_webhook:
        requeue_requires_input_requests(db)


def connection_status(
    connection_config: ConnectionConfig, msg: str, db: Session = Depends(deps.get_db)
) -> TestStatusMessage:
    """Connect, verify with a trivial query or API request, and report the status."""

    event_audit_service = EventAuditService(db)
    connection_service = ConnectionService(db, event_audit_service)
    return connection_service.connection_status(connection_config, msg)


def update_connection_secrets(
    connection_service: ConnectionService,
    connection_key: FidesKey,
    unvalidated_secrets: connection_secrets_schemas,
    verify: Optional[bool],
    merge_with_existing: Optional[bool] = False,
) -> TestStatusMessage:
    try:
        return connection_service.update_secrets(
            connection_key, unvalidated_secrets, verify, merge_with_existing
        )
    except ConnectionNotFoundException:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connection config found with key {connection_key}",
        )
    except SaaSConfigNotFoundException:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=validate_secrets_error_message(),
        )
    except ValidationError as e:
        errors = e.errors(include_url=False, include_input=False)
        for err in errors:
            # Additionally, manually remove the context from the error message -
            # this may contain sensitive information
            err.pop("ctx", None)
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(errors),
        )
    except FidesValidationError as e:
        # Check if the exception has the original pydantic errors attached
        raise HTTPException(status_code=HTTP_422_UNPROCESSABLE_ENTITY, detail=e.message)



from __future__ import annotations

import json
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Set

from httpx import AsyncClient
from loguru import logger
from sqlalchemy import text
from sqlalchemy.orm import Session
from sqlalchemy.sql.elements import TextClause

from fides.api.common_exceptions import PrivacyRequestError
from fides.api.models.privacy_request import (
    EXITED_EXECUTION_LOG_STATUSES,
    PrivacyRequest,
    RequestTask,
)
from fides.api.models.privacy_request.request_task import AsyncTaskType
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.drp_privacy_request import DrpPrivacyRequestCreate
from fides.api.schemas.policy import ActionType
from fides.api.schemas.privacy_request import PrivacyRequestStatus
from fides.api.schemas.redis_cache import Identity
from fides.api.tasks import DSR_QUEUE_NAME, DatabaseTask, celery_app
from fides.api.tasks.scheduled.scheduler import scheduler
from fides.api.util.cache import (
    FidesopsRedis,
    celery_tasks_in_flight,
    get_async_task_tracking_cache_key,
    get_cache,
    get_privacy_request_retry_count,
    increment_privacy_request_retry_count,
    reset_privacy_request_retry_count,
)
from fides.api.util.lock import redis_lock
from fides.config import CONFIG

PRIVACY_REQUEST_STATUS_CHANGE_POLL = "privacy_request_status_change_poll"
DSR_DATA_REMOVAL = "dsr_data_removal"
INTERRUPTED_TASK_REQUEUE_POLL = "interrupted_task_requeue_poll"
REQUEUE_INTERRUPTED_TASKS_LOCK = "requeue_interrupted_tasks_lock"
ASYNC_TASKS_STATUS_POLLING = "async_tasks_status_polling"
ASYNC_TASKS_STATUS_POLLING_LOCK = "async_tasks_status_polling_lock"
ASYNC_TASKS_STATUS_POLLING_LOCK_TIMEOUT = (
    300  # Starting timeout is shorter because the task goes directly to the workers.
)


def build_required_privacy_request_kwargs(
    requested_at: Optional[datetime],
    policy_id: str,
    verification_required: bool,
    authenticated: bool,
) -> Dict[str, Any]:
    """Build kwargs required for creating privacy request

    If identity verification is on, the request will have an initial status of
    "identity_unverified", otherwise, it will have an initial status of "pending".
    """
    status = (
        PrivacyRequestStatus.identity_unverified
        if verification_required and not authenticated
        else PrivacyRequestStatus.pending
    )
    return {
        "requested_at": requested_at,
        "policy_id": policy_id,
        "status": status,
    }


def cache_data(
    privacy_request: PrivacyRequest,
    identity: Identity,
    encryption_key: Optional[str],
    drp_request_body: Optional[DrpPrivacyRequestCreate],
    custom_privacy_request_fields: Optional[Dict[str, Any]] = None,
) -> None:
    """Cache privacy request data"""
    # Store identity and encryption key in the cache
    logger.info("Caching identity for privacy request {}", privacy_request.id)
    privacy_request.cache_identity(identity)
    privacy_request.cache_custom_privacy_request_fields(custom_privacy_request_fields)
    privacy_request.cache_encryption(encryption_key)  # handles None already

    if drp_request_body:
        privacy_request.cache_drp_request_body(drp_request_body)


def get_async_client() -> AsyncClient:
    """Return an async client used to make API requests"""
    return AsyncClient()


def initiate_poll_for_exited_privacy_request_tasks() -> None:
    """Initiates scheduler to check if a Privacy Request's status needs to be flipped when all
    Request Tasks have had a chance to run"""

    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add Privacy Request Status Change job."

    logger.info("Initiating scheduler for Privacy Request Status Change")
    scheduler.add_job(
        func=poll_for_exited_privacy_request_tasks,
        trigger="interval",
        kwargs={},
        id=PRIVACY_REQUEST_STATUS_CHANGE_POLL,
        coalesce=True,
        replace_existing=True,
        seconds=CONFIG.execution.state_polling_interval,
    )


@celery_app.task(base=DatabaseTask, bind=True)
def poll_for_exited_privacy_request_tasks(self: DatabaseTask) -> Set[str]:
    """
    Mark a privacy request as errored if all of its Request Tasks have run but some have errored.

    When a Request Task fails, it marks itself and *every Request Task that can be reached by the current
    Request Task* as failed. However, other Request Tasks independent of this path should still have an
    opportunity to run. We wait until everything has run before marking the Privacy Request as errored so it
    can be reprocessed.
    """
    with self.get_new_session() as db:
        logger.debug("Polling for privacy requests awaiting errored status change")

        # Privacy Requests that needed approval are in "Approved" status as they process.
        # Privacy Requests that didn't need approval are "In Processing".
        # Privacy Requests in these states should be examined to see if all of its Request Tasks have had a chance
        # to complete.
        # Use query_without_large_columns to prevent OOM errors when processing many privacy requests
        in_progress_privacy_requests = (
            PrivacyRequest.query_without_large_columns(db)
            .filter(
                PrivacyRequest.status.in_(
                    [
                        PrivacyRequestStatus.in_processing,
                        PrivacyRequestStatus.approved,
                        PrivacyRequestStatus.requires_input,
                    ]
                )
            )
            # Only look at Privacy Requests that haven't been deleted
            .filter(PrivacyRequest.deleted_at.is_(None))
            .order_by(PrivacyRequest.created_at)
        )

        # TODO: With this approach, we're making 3*n queries , where n is the number of in-progress privacy requests.
        # We could optimize this to just get all the RequestTasks for all in-progress privacy requests in one single
        # query and then process them in-memory. This could be more efficient.
        def privacy_request_has_errored_tasks(
            privacy_request: PrivacyRequest, task_type: ActionType
        ) -> bool:
            """Check if a privacy request has exited all its tasks of the given type,
            and at least one of the tasks has errored.
            We specifically only query for the RequestTask.status column to avoid
            querying for the entire RequestTask row.
            """
            tasks_statuses_query = db.query(RequestTask.status).filter(
                RequestTask.privacy_request_id == privacy_request.id,
                RequestTask.action_type == task_type,
            )

            statuses = set(status for status, in tasks_statuses_query.all())
            all_exited = all(
                status in EXITED_EXECUTION_LOG_STATUSES for status in statuses
            )
            if all_exited and ExecutionLogStatus.error in statuses:
                logger.info(
                    f"Marking {task_type.value} step of {privacy_request.id} as error"
                )
                return True

            return False

        marked_as_errored: Set[str] = set()
        for pr in in_progress_privacy_requests.all():
            if pr.consent_tasks.count():
                # Consent propagation tasks - these are not created until access and erasure steps are complete.
                if privacy_request_has_errored_tasks(pr, ActionType.consent):
                    pr.error_processing(db)
                    marked_as_errored.add(pr.id)

            if pr.erasure_tasks.count():
                # Erasure tasks are created at the same time as access tasks but if any are errored, this means
                # we made it to the erasure section
                if privacy_request_has_errored_tasks(pr, ActionType.erasure):
                    pr.error_processing(db)
                    marked_as_errored.add(pr.id)

            if pr.access_tasks.count():
                if privacy_request_has_errored_tasks(pr, ActionType.access):
                    pr.error_processing(db)
                    marked_as_errored.add(pr.id)

        return marked_as_errored


def initiate_scheduled_dsr_data_removal() -> None:
    """Initiates scheduler to cleanup obsolete access and erasure data"""

    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add DSR data removal job."

    logger.info("Initiating scheduler for DSR Data Removal")
    scheduler.add_job(
        func=remove_saved_dsr_data,
        kwargs={},
        id=DSR_DATA_REMOVAL,
        coalesce=False,
        replace_existing=True,
        trigger="cron",
        minute="0",
        hour="2",
        day="*",
        timezone="US/Eastern",
    )


@celery_app.task(base=DatabaseTask, bind=True)
def remove_saved_dsr_data(self: DatabaseTask) -> None:
    """
    Remove saved customer data that is no longer needed to facilitate running the access or erasure request.
    """
    with self.get_new_session() as db:
        logger.info("Running DSR Data Removal Task to cleanup obsolete user data")

        # Remove old request tasks which potentially contain encrypted PII
        remove_dsr_data: TextClause = text(
            """
            DELETE FROM requesttask
            USING privacyrequest
            WHERE requesttask.privacy_request_id = privacyrequest.id
            AND requesttask.created_at < :ttl
            AND privacyrequest.status = 'complete';
            """
        )

        result = db.execute(
            remove_dsr_data,
            {
                "ttl": (
                    datetime.now()
                    - timedelta(seconds=CONFIG.execution.request_task_ttl)
                ),
            },
        )
        affected_rows = result.rowcount
        logger.info(
            f"Deleted {affected_rows} expired request tasks via DSR Data Removal Task."
        )

        # Remove columns from old privacyrequests that potentially contain encrypted PII
        # or URL's that contain encrypted PII.
        remove_data_from_privacy_request: TextClause = text(
            """
            UPDATE privacyrequest
            SET filtered_final_upload = null, access_result_urls = null
            WHERE privacyrequest.updated_at < :ttl
            AND privacyrequest.status = 'complete';
            """
        )

        db.execute(
            remove_data_from_privacy_request,
            {
                "ttl": (  # Using Redis Default TTL Seconds by default
                    datetime.now() - timedelta(seconds=CONFIG.redis.default_ttl_seconds)
                ),
            },
        )

        db.commit()


def initiate_interrupted_task_requeue_poll() -> None:
    """Initiates scheduler to check for and requeue interrupted tasks"""

    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add interrupted task requeue job."

    logger.info("Initiating scheduler for interrupted task requeue")
    scheduler.add_job(
        func=requeue_interrupted_tasks,
        trigger="interval",
        kwargs={},
        id=INTERRUPTED_TASK_REQUEUE_POLL,
        coalesce=True,
        replace_existing=True,
        seconds=CONFIG.execution.interrupted_task_requeue_interval,
    )


def initiate_polling_task_requeue() -> None:
    """Initiates scheduler to check for and requeue pending polling async tasks"""
    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot add async tasks status polling job."

    logger.info("Initiating scheduler for async tasks status polling")
    scheduler.add_job(
        func=requeue_polling_tasks,
        trigger="interval",
        kwargs={},
        id=ASYNC_TASKS_STATUS_POLLING,
        coalesce=True,
        replace_existing=True,
        seconds=CONFIG.execution.async_polling_interval_hours * 3600,
    )


def get_cached_task_id(entity_id: str) -> Optional[str]:
    """Gets the cached task ID for a privacy request or request task by ID.

    Raises Exception if cache operations fail, allowing callers to handle cache failures appropriately.
    """
    cache: FidesopsRedis = get_cache()
    try:
        task_id = cache.get(get_async_task_tracking_cache_key(entity_id))
        return task_id
    except Exception as exc:
        logger.error(f"Failed to get cached task ID for entity {entity_id}: {exc}")
        raise


def _get_task_ids_from_dsr_queue(
    redis_client: FidesopsRedis, chunk_size: int = 100
) -> Set[str]:
    """
    Get all task IDs from a Redis queue in a memory-efficient way.

    Args:
        redis_client: Redis client
        chunk_size: Size of chunks to process

    Returns:
        Set of task IDs
    """
    queued_tasks_ids = set()
    queue_length = redis_client.llen(DSR_QUEUE_NAME)

    # Process the queue in chunks to avoid memory issues with large queues
    for offset in range(0, queue_length, chunk_size):
        # Get a chunk of tasks (at most chunk_size elements)
        tasks_chunk = redis_client.lrange(
            DSR_QUEUE_NAME, offset, offset + chunk_size - 1
        )

        # Extract task IDs from the chunk
        for task in tasks_chunk:
            try:
                queued_tasks_ids.add(json.loads(task)["headers"]["id"])
            except (json.JSONDecodeError, KeyError) as e:
                logger.warning(f"Failed to parse task from queue: {e}")
                continue

    return queued_tasks_ids


def _cancel_interrupted_tasks_and_error_privacy_request(
    db: Session, privacy_request: PrivacyRequest, error_message: Optional[str] = None
) -> None:
    """
    Cancel all tasks associated with an interrupted privacy request and set the privacy request to error state.

    This function:
    1. Logs the error message (either provided or default)
    2. Revokes the main privacy request task and all associated request tasks
    3. Sets the privacy request status to error
    4. Creates an error log entry

    Args:
        db: Database session
        privacy_request: The privacy request to cancel and error
        error_message: Optional error message to log. If not provided, uses default message.
    """
    if error_message:
        logger.error(error_message)
    else:
        logger.error(
            f"Canceling interrupted tasks and marking privacy request {privacy_request.id} as error"
        )

    # Cancel all associated Celery tasks
    privacy_request.cancel_celery_tasks()

    # Set privacy request to error state using the existing method
    try:
        privacy_request.error_processing(db)
        logger.info(
            f"Privacy request {privacy_request.id} marked as error due to task interruption"
        )
    except Exception as exc:
        logger.error(
            f"Failed to mark privacy request {privacy_request.id} as error: {exc}"
        )


def _handle_privacy_request_requeue(
    db: Session, privacy_request: PrivacyRequest
) -> None:
    """Handle retry logic for a privacy request - either requeue or cancel based on retry count."""
    try:
        # Check retry count and either requeue or cancel based on limit
        current_retry_count = get_privacy_request_retry_count(privacy_request.id)
        max_retries = CONFIG.execution.privacy_request_requeue_retry_count

        if current_retry_count < max_retries:
            # Increment retry count and attempt requeue
            new_retry_count = increment_privacy_request_retry_count(privacy_request.id)
            logger.info(
                f"Requeuing privacy request {privacy_request.id} "
                f"(attempt {new_retry_count}/{max_retries})"
            )

            from fides.service.privacy_request.privacy_request_service import (  # pylint: disable=cyclic-import
                _requeue_privacy_request,
            )

            try:
                _requeue_privacy_request(db, privacy_request)
            except PrivacyRequestError as exc:
                # If requeue fails, cancel tasks and set to error state
                _cancel_interrupted_tasks_and_error_privacy_request(
                    db, privacy_request, exc.message
                )
        else:
            # Exceeded retry limit, cancel tasks and set to error state
            _cancel_interrupted_tasks_and_error_privacy_request(
                db,
                privacy_request,
                f"Privacy request {privacy_request.id} exceeded max retry attempts "
                f"({max_retries}), canceling tasks and setting to error state",
            )
            # Reset retry count since we're giving up
            reset_privacy_request_retry_count(privacy_request.id)

    except Exception as cache_exc:
        # If cache operations fail (Redis down, network issues, etc.), fail safe by canceling
        _cancel_interrupted_tasks_and_error_privacy_request(
            db,
            privacy_request,
            f"Cache operation failed for privacy request {privacy_request.id}, "
            f"failing safe by canceling tasks: {cache_exc}",
        )


def _get_request_task_ids_in_progress(
    db: Session, privacy_request_id: str
) -> List[str]:
    """Get the IDs of request tasks that are currently in progress for a privacy request."""
    request_tasks_in_progress = (
        db.query(RequestTask.id)
        .filter(RequestTask.privacy_request_id == privacy_request_id)
        .filter(
            RequestTask.status.in_(
                [
                    ExecutionLogStatus.in_processing,
                    ExecutionLogStatus.pending,
                ]
            )
        )
        .all()
    )
    return [task[0] for task in request_tasks_in_progress]


def _has_polling_tasks(db: Session, privacy_request_id: str) -> bool:
    """
    Check if a privacy request has any polling async tasks.

    Args:
        db: Database session
        privacy_request_id: The ID of the privacy request to check

    Returns:
        bool: True if the privacy request has polling tasks, False otherwise
    """
    return db.query(
        db.query(RequestTask)
        .filter(
            RequestTask.privacy_request_id == privacy_request_id,
            RequestTask.async_type == AsyncTaskType.polling,
        )
        .exists()
    ).scalar()


# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
@celery_app.task(base=DatabaseTask, bind=True)
def requeue_interrupted_tasks(self: DatabaseTask) -> None:
    """
    Requeue interrupted tasks for privacy requests that are in progress.

    This function checks for privacy requests that are currently pending or in progress and
    verifies if their associated tasks are still in the queue or running. If any
    task is found to be interrupted (i.e., not in the queue and not running), the
    privacy request is requeued to ensure its completion.

    The function performs the following steps:
    1. Retrieves all task IDs currently in the queue.
    2. Fetches all in-progress privacy requests from the database.
    3. Checks each privacy request to determine if its tasks are still active.
    4. Requeues the privacy request if any of its tasks are found to be interrupted.
    """
    redis_conn: FidesopsRedis = get_cache()

    with redis_lock(REQUEUE_INTERRUPTED_TASKS_LOCK, 600) as lock:
        if not lock:
            return
        with self.get_new_session() as db:
            logger.debug("Starting check for interrupted tasks to requeue")

            # Get all in-progress privacy requests that haven't been updated in the last 5 minutes
            # Use query_without_large_columns to prevent OOM errors when processing many privacy requests
            in_progress_requests = (
                PrivacyRequest.query_without_large_columns(db)
                .filter(
                    PrivacyRequest.status.in_(
                        [
                            PrivacyRequestStatus.in_processing,
                            PrivacyRequestStatus.approved,
                            PrivacyRequestStatus.requires_input,
                        ]
                    )
                )
                .filter(PrivacyRequest.deleted_at.is_(None))
                .order_by(PrivacyRequest.created_at)
            )

            if not in_progress_requests.count():
                logger.debug("No in-progress privacy requests to check")
                return

            logger.debug(
                f"Found {in_progress_requests.count()} privacy requests to check"
            )

            # Get task IDs from the queue in a memory-efficient way
            try:
                queued_tasks_ids = _get_task_ids_from_dsr_queue(redis_conn)
            except Exception as queue_exc:
                logger.warning(
                    f"Failed to get task IDs from queue, skipping queue state checks: {queue_exc}"
                )
                return

            # Check each privacy request
            for privacy_request in in_progress_requests:
                should_requeue = False
                logger.debug(f"Checking tasks for privacy request {privacy_request.id}")

                try:
                    task_id = get_cached_task_id(privacy_request.id)
                except Exception as cache_exc:
                    # If we can't get the task ID due to cache failure, fail safe by canceling
                    _cancel_interrupted_tasks_and_error_privacy_request(
                        db,
                        privacy_request,
                        f"Cache failure when getting task ID for privacy request {privacy_request.id}, "
                        f"failing safe by canceling tasks: {cache_exc}",
                    )
                    continue

                # If the task ID is not cached, we can't check if it's running
                # This means the request is stuck - cancel it
                if not task_id:
                    _cancel_interrupted_tasks_and_error_privacy_request(
                        db,
                        privacy_request,
                        f"No task ID found for privacy request {privacy_request.id}, "
                        f"request is stuck without a running task - canceling",
                    )
                    continue

                # Check if the main privacy request task is active
                if task_id not in queued_tasks_ids and not celery_tasks_in_flight(
                    [task_id]
                ):
                    request_tasks_count = (
                        db.query(RequestTask)
                        .filter(RequestTask.privacy_request_id == privacy_request.id)
                        .count()
                    )
                    if request_tasks_count == 0:
                        logger.warning(
                            f"The task for privacy request {privacy_request.id} was terminated before it could schedule any request tasks, requeueing privacy request"
                        )
                        should_requeue = True

                    request_task_ids_in_progress = _get_request_task_ids_in_progress(
                        db, privacy_request.id
                    )

                    # Check each individual request task
                    for request_task_id in request_task_ids_in_progress:
                        try:
                            subtask_id = get_cached_task_id(request_task_id)
                        except Exception as cache_exc:
                            # If we can't get the subtask ID due to cache failure, fail safe by canceling
                            _cancel_interrupted_tasks_and_error_privacy_request(
                                db,
                                privacy_request,
                                f"Cache failure when getting subtask ID for request task {request_task_id} "
                                f"(privacy request {privacy_request.id}), failing safe by canceling tasks: {cache_exc}",
                            )
                            should_requeue = False
                            break

                        # If the task ID is not cached, we can't check if it's running
                        # This means the subtask is stuck - but we need to handle this differently
                        # based on the privacy request status
                        if not subtask_id:
                            if (
                                privacy_request.status
                                == PrivacyRequestStatus.requires_input
                            ):
                                # For requires_input status, don't automatically error the request
                                # as it's intentionally waiting for user input
                                logger.warning(
                                    f"No task ID found for request task {request_task_id} "
                                    f"(privacy request {privacy_request.id}) in requires_input status - "
                                    f"keeping request in current status as it may be waiting for manual input"
                                )
                                should_requeue = False
                                break

                            # Check if the Privacy request has polling tasks
                            if _has_polling_tasks(db, privacy_request.id):
                                # If the polling request task has no cached task ID, it's stuck
                                logger.warning(
                                    f"No task ID found for request task {request_task_id} "
                                    f"(privacy request {privacy_request.id}) Contains polling tasks - "
                                    f"keeping request in current status as it may be waiting for polling task to complete"
                                )
                                should_requeue = False
                                break

                            # For other statuses, cancel the entire privacy request
                            _cancel_interrupted_tasks_and_error_privacy_request(
                                db,
                                privacy_request,
                                f"No task ID found for request task {request_task_id} "
                                f"(privacy request {privacy_request.id}), subtask is stuck - canceling privacy request",
                            )
                            should_requeue = False
                            break

                        if (
                            subtask_id not in queued_tasks_ids
                            and not celery_tasks_in_flight([subtask_id])
                        ):
                            logger.warning(
                                f"Request task {request_task_id} is not in the queue or running, requeueing privacy request"
                            )
                            should_requeue = True
                            break

                # Requeue the privacy request if needed
                if should_requeue:
                    _handle_privacy_request_requeue(db, privacy_request)


@celery_app.task(base=DatabaseTask, bind=True)
def requeue_polling_tasks(self: DatabaseTask) -> None:
    """
    Poll the status of async tasks that are awaiting processing.
    """
    with redis_lock(
        ASYNC_TASKS_STATUS_POLLING_LOCK,
        ASYNC_TASKS_STATUS_POLLING_LOCK_TIMEOUT,
    ) as lock:
        if not lock:
            logger.debug("Async tasks status polling lock not acquired, skipping")
            return
        with self.get_new_session() as db:
            logger.debug("Polling for async tasks status")

            # Get all tasks that are polling and are from polling async tasks
            async_tasks = (
                db.query(RequestTask)
                .filter(RequestTask.status == ExecutionLogStatus.polling)
                .filter(RequestTask.async_type == AsyncTaskType.polling)
                .all()
            )
            logger.info(f"Found {len(async_tasks)} async polling tasks")

            if async_tasks:
                # Avoiding cyclic imports
                from fides.api.task.execute_request_tasks import queue_request_task

                for async_task in async_tasks:
                    logger.info(
                        f"Requeuing polling task {async_task.id} for processing"
                    )
                    queue_request_task(async_task, privacy_request_proceed=True)



import types
import uuid
from typing import Any, Optional

from pydantic import BaseModel, Field
from sqlalchemy import Column, any_, bindparam

from fides.api.task.conditional_dependencies.schemas import Operator


class SQLTranslationError(Exception):
    """Error raised when SQL translation fails"""


def _escape_like_pattern(val: Any) -> str:
    """
    Escape LIKE wildcards in user input to prevent pattern injection attacks.

    This prevents users from injecting wildcards (% and _) that could:
    - Match unintended records through pattern injection
    - Cause performance issues with expensive wildcard operations
    - Enable information disclosure through pattern probing

    Note: This function escapes ALL % and _ characters as a security measure.
    While this may seem aggressive for normal field names, it's necessary because:
    1. We can't distinguish between intentional and malicious wildcards
    2. Field values shouldn't typically contain SQL wildcards
    3. Better to be overly cautious with user input

    Args:
        val: User input value to escape

    Returns:
        Escaped string safe for use in LIKE patterns

    Examples:
        _escape_like_pattern("admin%") -> "admin\\%"
        _escape_like_pattern("test_user") -> "test\\_user"  # Escapes _ for security
        _escape_like_pattern("normal") -> "normal"
    """
    if val is None:
        return ""

    val_str = str(val)
    # Escape backslashes first to prevent double-escaping
    # Then escape LIKE wildcards (% and _)
    return val_str.replace("\\", "\\\\").replace("%", "\\%").replace("_", "\\_")


def _validate_and_escape_json_path_component(component: str) -> str:
    """
    Validate and escape JSON path components to prevent injection attacks.

    This prevents users from injecting malicious content in JSON path components that could:
    - Break out of single quotes in PostgreSQL JSON operators
    - Inject arbitrary SQL through malformed JSON paths
    - Cause syntax errors or unexpected behavior

    Args:
        component: JSON path component to validate and escape

    Returns:
        Validated and escaped component safe for use in PostgreSQL JSON operators

    Raises:
        SQLTranslationError: If the component contains invalid characters

    Examples:
        _validate_and_escape_json_path_component("field") -> "field"
        _validate_and_escape_json_path_component("field'name") -> "field''name"
        _validate_and_escape_json_path_component("") -> raises SQLTranslationError
    """
    if not component or not isinstance(component, str):
        raise SQLTranslationError(
            f"Invalid JSON path component: '{component}'. Components must be non-empty strings."
        )

    # Check for potentially dangerous characters
    if len(component) > 100:  # Reasonable limit for JSON field names
        raise SQLTranslationError(
            f"JSON path component too long: '{component[:50]}...'. Maximum length is 100 characters."
        )

    # Escape single quotes for PostgreSQL JSON operators (double them)
    escaped_component = component.replace("'", "''")

    return escaped_component


def _handle_list_contains(col: Column, val: Any) -> Column:
    """
    Handle list_contains operator for different scenarios:
    - If val is a list: check if column value is IN the list
    - If val is a single value: check if column contains the value (for arrays/JSON) or use LIKE (for strings)

    Args:
        col: SQLAlchemy column
        val: Value to compare against

    Returns:
        SQLAlchemy expression for the comparison
    """
    if isinstance(val, list):
        # If value is a list, check if column value is IN the list
        return col.in_(val)

    # For single values, we need to handle different column types
    # Check if this is a PostgreSQL array column
    if hasattr(col.type, "item_type") or str(col.type).startswith("ARRAY"):
        # This is a PostgreSQL array - use the ANY operator with proper parameter binding
        # This is safer and prevents SQL injection
        # Generate unique parameter name to avoid conflicts when multiple list operations exist in same query
        unique_param_name = f"array_val_{uuid.uuid4().hex[:8]}"
        param = bindparam(unique_param_name, val)
        return param == any_(col)

    # Try JSON containment for JSONB/JSON columns
    try:
        return col.contains(val)
    except Exception:
        # Fallback to LIKE for string columns with escaped wildcards
        escaped_val = _escape_like_pattern(val)
        return col.like(f"%{escaped_val}%", escape="\\")


OPERATOR_MAP = types.MappingProxyType(
    {
        Operator.eq: lambda col, val: col == val,
        Operator.neq: lambda col, val: col != val,
        Operator.lt: lambda col, val: col < val,
        Operator.lte: lambda col, val: col <= val,
        Operator.gt: lambda col, val: col > val,
        Operator.gte: lambda col, val: col >= val,
        Operator.contains: lambda col, val: col.like(
            f"%{_escape_like_pattern(val)}%", escape="\\"
        ),
        Operator.starts_with: lambda col, val: col.like(
            f"{_escape_like_pattern(val)}%", escape="\\"
        ),
        Operator.ends_with: lambda col, val: col.like(
            f"%{_escape_like_pattern(val)}", escape="\\"
        ),
        Operator.exists: lambda col, val: col.isnot(None),
        Operator.not_exists: lambda col, val: col.is_(None),
        Operator.list_contains: _handle_list_contains,
        Operator.not_in_list: lambda col, val: ~_handle_list_contains(col, val),
    }
)


class FieldAddress(BaseModel):
    """Parsed field address with table and column information"""

    table_name: str = Field(description="Table name extracted from field address")
    column_name: str = Field(description="Base column name without JSON path")
    json_path: Optional[list[str]] = Field(
        default=None, description="JSON path components if this is a JSON field"
    )
    full_address: str = Field(description="Original field address string")

    def __hash__(self) -> int:
        """Make FieldAddress hashable for use in sets"""
        return hash(
            (
                self.table_name,
                self.column_name,
                tuple(self.json_path) if self.json_path else None,
            )
        )

    def __eq__(self, other: Any) -> bool:
        """Enable equality comparison for FieldAddress objects"""
        if not isinstance(other, FieldAddress):
            return False
        return (
            self.table_name == other.table_name
            and self.column_name == other.column_name
            and self.json_path == other.json_path
        )

    def to_sql_column(self, enable_json_operators: bool = True) -> str:
        """
        Convert field address to PostgreSQL column reference

        Args:
            enable_json_operators: Whether to use PostgreSQL JSON operators for JSON paths

        Returns:
            SQL column reference string
        """
        is_json_path = self.json_path is not None and len(self.json_path) > 0
        if not is_json_path or not enable_json_operators:
            return self.column_name

        if self.json_path is None:
            # This should never happen
            raise SQLTranslationError(
                "Field address internal error."
            )  # pragma: no cover

        # Build PostgreSQL JSON path: column->'path'->'path'->>'final_path'
        # Validate and escape all components to prevent injection
        if self.json_path and len(self.json_path) == 1:
            # Simple case: column->>'field'
            escaped_component = _validate_and_escape_json_path_component(
                self.json_path[0]
            )
            return f"{self.column_name}->>'{escaped_component}'"

        # Complex case: column->'field'->'field'->>'final_field'
        path_parts = []
        # All but last use -> operator
        for part in self.json_path[:-1]:
            escaped_part = _validate_and_escape_json_path_component(part)
            path_parts.append(f"->'{escaped_part}'")
        # Last part uses ->> operator (returns text)
        escaped_final = _validate_and_escape_json_path_component(self.json_path[-1])
        path_parts.append(f"->>'{escaped_final}'")

        return f"{self.column_name}{''.join(path_parts)}"

    @classmethod
    def _parse_parts_to_components(
        cls, parts: list[str]
    ) -> tuple[str, str, Optional[list[str]]]:
        """
        Parse a list of parts into table_name, column_name, and json_path components.

        Args:
            parts: List of address parts (e.g., ["table", "column", "path1", "path2"])

        Returns:
            Tuple of (table_name, column_name, json_path)
        """
        if len(parts) < 2:
            return "", parts[0] if parts else "", None

        table_name = parts[0]
        column_name = parts[1]
        json_path = parts[2:] if len(parts) > 2 else None

        return table_name, column_name, json_path

    @classmethod
    def parse(cls, field_address: str) -> "FieldAddress":
        """
        Parse a field address into components

        Supports formats:
        - "table.column" -> table="table", column="column"
        - "table.json_column.path.subpath" -> table="table", column="json_column", json_path=["path", "subpath"]
        - "table:column" -> table="table", column="column" (alternative format)
        - "column" -> table="", column="column" (requires default table)
        """
        if ":" in field_address:
            # Format: table:column or dataset:collection:field
            # Handle colon-separated format first, then check for dots in the remaining parts
            colon_parts = field_address.split(":")
            if len(colon_parts) >= 2:
                table_name = colon_parts[0]
                remaining = ":".join(colon_parts[1:])  # Join remaining parts

                # Check if the remaining part has dots (JSON path)
                if "." in remaining:
                    # Mixed format: table:column.path1.path2
                    dot_parts = remaining.split(".")
                    parts = [table_name] + dot_parts
                else:
                    # Pure colon format: table:column or table:column:path1:path2
                    parts = colon_parts

                table_name, column_name, json_path = cls._parse_parts_to_components(
                    parts
                )

                return cls(
                    table_name=table_name,
                    column_name=column_name,
                    json_path=json_path,
                    full_address=field_address,
                )

        if "." in field_address:
            # Format: table.column or table.json_column.path.subpath
            parts = field_address.split(".")
            table_name, column_name, json_path = cls._parse_parts_to_components(parts)

            return cls(
                table_name=table_name,
                column_name=column_name,
                json_path=json_path,
                full_address=field_address,
            )

        # Simple column name without table
        return cls(
            table_name="",  # Will need default table
            column_name=field_address,
            json_path=None,
            full_address=field_address,
        )



from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, Union

from sqlalchemy import Column, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base
from fides.api.db.util import EnumColumn
from fides.api.task.conditional_dependencies.schemas import (
    Condition,
    ConditionGroup,
    ConditionLeaf,
)

if TYPE_CHECKING:
    from sqlalchemy.orm.relationships import RelationshipProperty


class ConditionalDependencyError(Exception):
    """Exception for conditional dependency errors."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class ConditionalDependencyType(str, Enum):
    """Shared enum for conditional dependency node types.

    Attributes:
        leaf: Individual condition (field_address + operator + value)
        group: Collection of conditions with logical operator (AND/OR)
    """

    leaf = "leaf"
    group = "group"


class ConditionalDependencyBase(Base):
    """Abstract base class for all conditional dependency models.

    This class provides a common structure for building hierarchical condition trees
    that can be evaluated to determine when certain actions should be taken.

    Architecture:
        - Tree Structure: Supports parent-child relationships for complex logic
        - Two Node Types: 'leaf' (individual conditions) and 'group' (logical operators)
        - Flexible Schema: Uses JSONB for dynamic value storage
        - Ordered Evaluation: sort_order ensures predictable condition processing

    Concrete Implementations:
        - ManualTaskConditionalDependency: Single-type hierarchy for manual tasks
          - Single-type hierarchy means one condition tree per manual task, this condition
            may be a nested group of conditions or a single leaf condition.
        - DigestCondition: Multi-type hierarchy with digest_condition_type separation
          - Multi-type hierarchy means one digest_config can have multiple independent
            condition trees, each with a different digest_condition_type (RECEIVER, CONTENT, PRIORITY)
          - Within each tree, all nodes must have the same digest_condition_type
          - This enables separate condition logic for different aspects of digest processing

    Usage Pattern:
        1. Inherit from this base class
        2. Define your table name with @declared_attr
        3. Add foreign key relationships (parent_id, entity_id)
        4. Implement get_root_condition() classmethod
        5. Add any domain-specific columns

    Example Tree Structure:
        Root Group (AND)
        ├── Leaf: user.role == "admin"
        ├── Leaf: request.priority >= 3
        └── Child Group (OR)
            ├── Leaf: user.department == "security"
            └── Leaf: user.department == "compliance"

    Note:
        - This is a SQLAlchemy abstract model (__abstract__ = True)
        - No database table is created for this base class
        - Subclasses must implement get_root_condition()
        - The 'children' relationship must be defined in concrete subclasses
    """

    __abstract__ = True

    # Tree structure - parent_id defined in concrete classes for proper foreign keys
    condition_type = Column(
        EnumColumn(ConditionalDependencyType), nullable=False, index=True
    )

    # Condition details (for leaf nodes)
    field_address = Column(String(255), nullable=True)  # For leaf conditions
    operator = Column(String, nullable=True)  # For leaf conditions
    value = Column(JSONB, nullable=True)  # For leaf conditions
    logical_operator = Column(String, nullable=True)  # 'and' or 'or' for groups

    # Ordering
    sort_order = Column(Integer, nullable=False, default=0, index=True)

    def to_correct_condition_type(self) -> Union[ConditionLeaf, ConditionGroup]:
        """Convert this database model to the correct condition type."""
        if self.condition_type == ConditionalDependencyType.leaf:
            return self.to_condition_leaf()
        return self.to_condition_group()

    def to_condition_leaf(self) -> ConditionLeaf:
        """Convert this database model to a ConditionLeaf schema object.

        This method transforms a leaf-type conditional dependency from its database
        representation into a structured ConditionLeaf object that can be used for
        evaluation and serialization.

        Returns:
            ConditionLeaf: Schema object containing field_address, operator, and value

        Raises:
            ValueError: If this condition is not a leaf type (i.e., it's a group)

        Example:
            >>> condition = SomeConcreteConditionalDependency(
            ...     condition_type="leaf",
            ...     field_address="user.role",
            ...     operator="eq",
            ...     value="admin"
            ... )
            >>> leaf = condition.to_condition_leaf()
            >>> print(leaf.field_address)  # "user.role"
        """
        if self.condition_type != ConditionalDependencyType.leaf:
            raise ValueError(
                f"Cannot convert {self.condition_type} condition to leaf. "
                f"Only conditions with condition_type='leaf' can be converted to ConditionLeaf. "
                f"This condition has type '{self.condition_type}' and should be converted using to_condition_group()."
            )

        return ConditionLeaf(
            field_address=self.field_address, operator=self.operator, value=self.value
        )

    def to_condition_group(self) -> ConditionGroup:
        """Convert this database model to a ConditionGroup schema object.

        This method transforms a group-type conditional dependency from its database
        representation into a structured ConditionGroup object. It recursively processes
        all child conditions, maintaining the tree structure and sort order.

        Returns:
            ConditionGroup: Schema object containing logical_operator and child conditions

        Raises:
            ValueError: If this condition is not a group type (i.e., it's a leaf)
            AttributeError: If the 'children' relationship is not properly defined

        Example:
            >>> # Assume we have a group with two leaf children
            >>> group_condition = SomeConcreteConditionalDependency(
            ...     condition_type="group",
            ...     logical_operator="and"
            ... )
            >>> condition_group = group_condition.to_condition_group()
            >>> print(condition_group.logical_operator)  # "and"
            >>> print(len(condition_group.conditions))   # 2
        """
        if self.condition_type != ConditionalDependencyType.group:
            raise ValueError(
                f"Cannot convert {self.condition_type} condition to group. "
                f"Only conditions with condition_type='group' can be converted to ConditionGroup. "
                f"This condition has type '{self.condition_type}' and should be converted using to_condition_leaf()."
            )

        # Recursively build children - note: 'children' must be defined in concrete classes
        try:
            children_list = [child for child in self.children]  # type: ignore[attr-defined]
        except AttributeError:
            raise AttributeError(
                f"The 'children' relationship is not defined on {self.__class__.__name__}. "
                f"Concrete subclasses must define a 'children' relationship for group conditions to work properly."
            )

        child_conditions = []
        for child in sorted(children_list, key=lambda x: x.sort_order):
            if child.condition_type == ConditionalDependencyType.leaf:
                child_conditions.append(child.to_condition_leaf())
            elif child.condition_type == ConditionalDependencyType.group:
                child_conditions.append(child.to_condition_group())
            else:
                raise ValueError(
                    f"Unknown condition_type '{child.condition_type}' found in child condition. "
                    f"Expected '{ConditionalDependencyType.leaf}' or '{ConditionalDependencyType.group}'."
                )

        return ConditionGroup(
            logical_operator=self.logical_operator, conditions=child_conditions
        )

    @classmethod
    def get_root_condition(cls, db: Session, **kwargs: Any) -> Optional[Condition]:
        """Get the root condition tree for a parent entity.

        This abstract method must be implemented by concrete subclasses to define
        how to retrieve the root condition node for their specific use case.
        The root condition represents the top-level node in a condition tree.

        Implementation Guidelines:
            1. Query for conditions with parent_id=None for the given parent entity
            2. Return None if no root condition exists
            3. Convert the database model to a Condition schema object
            4. Handle any domain-specific filtering or validation

        Args:
            db: SQLAlchemy database session for querying
            **kwargs: Keyword arguments specific to each implementation.
                    Examples:
                    - manual_task_id: ID of the manual task (for single-type hierarchies)
                    - digest_config_id: ID of the digest config (for multi-type hierarchies)
                    - digest_condition_type: Type of digest condition (for multi-type hierarchies)

        Returns:
            Optional[Condition]: Root condition tree (ConditionLeaf or ConditionGroup) or None
                               if no conditions exist for the specified criteria

        Raises:
            NotImplementedError: If called on the base class directly

        Example Implementation:
            >>> @classmethod
            >>> def get_root_condition(cls, db: Session, *, manual_task_id: str) -> Optional[Condition]:
            ...     root = db.query(cls).filter(
            ...         cls.manual_task_id == manual_task_id,
            ...         cls.parent_id.is_(None)
            ...     ).first()
            ...     if not root:
            ...         return None
            ...     return root.to_condition_leaf() if root.condition_type == 'leaf' else root.to_condition_group()
        """
        raise NotImplementedError(
            f"Subclasses of {cls.__name__} must implement get_root_condition(). "
            f"This method should query for the root condition (parent_id=None) "
            f"and return it as a Condition schema object, or None if not found. "
            f"See the docstring for implementation guidelines and examples."
        )

    def get_depth(self) -> int:
        """Calculate the depth of this node in the condition tree.

        Returns:
            int: Depth level (0 for root, 1 for direct children, etc.)

        Note:
            Requires the 'parent' relationship to be defined in concrete classes.
        """
        depth = 0
        current = self
        try:
            while hasattr(current, "parent") and current.parent is not None:  # type: ignore[attr-defined]
                depth += 1
                current = current.parent  # type: ignore[attr-defined]
        except AttributeError:
            # If parent relationship not defined, we can't calculate depth
            pass
        return depth

    def get_tree_summary(self) -> str:
        """Generate a human-readable summary of this condition tree.

        Returns:
            str: Multi-line string representation of the condition tree structure

        Example:
            >>> print(condition.get_tree_summary())
            Group (AND) [depth: 0, order: 0]
            ├── Leaf: user.role == "admin" [depth: 1, order: 0]
            ├── Leaf: request.priority >= 3 [depth: 1, order: 1]
            └── Group (OR) [depth: 1, order: 2]
                ├── Leaf: user.dept == "security" [depth: 2, order: 0]
                └── Leaf: user.dept == "compliance" [depth: 2, order: 1]
        """

        def _build_tree_lines(
            node: "ConditionalDependencyBase", prefix: str = "", is_last: bool = True
        ) -> list[str]:
            lines = []

            # Current node info
            if node.condition_type == ConditionalDependencyType.leaf:
                node_desc = f"Leaf: {node.field_address} {node.operator} {node.value}"
            else:
                node_desc = f"Group ({node.logical_operator.upper() if node.logical_operator else 'UNKNOWN'})"

            depth = node.get_depth()
            connector = "└── " if is_last else "├── "
            lines.append(
                f"{prefix}{connector}{node_desc} [depth: {depth}, order: {node.sort_order}]"
            )

            # Add children if this is a group
            if node.condition_type == ConditionalDependencyType.group:
                try:
                    children = sorted([child for child in node.children], key=lambda x: x.sort_order)  # type: ignore[attr-defined]
                    for i, child in enumerate(children):
                        is_last_child = i == len(children) - 1
                        child_prefix = prefix + ("    " if is_last else "│   ")
                        lines.extend(
                            _build_tree_lines(child, child_prefix, is_last_child)
                        )
                except AttributeError:
                    lines.append(f"{prefix}    [children relationship not defined]")

            return lines

        return "\n".join(_build_tree_lines(self))



from enum import Enum
from typing import TYPE_CHECKING, Any, Optional, Union

from sqlalchemy import Column, ForeignKey, Index, String, text
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship

from fides.api.db.base_class import FidesBase
from fides.api.db.util import EnumColumn
from fides.api.models.conditional_dependency.conditional_dependency_base import (
    ConditionalDependencyBase,
    ConditionalDependencyError,
)
from fides.api.task.conditional_dependencies.schemas import (
    Condition,
    ConditionGroup,
    ConditionLeaf,
)

if TYPE_CHECKING:
    from fides.api.models.digest.digest_config import DigestConfig


class DigestConditionType(str, Enum):
    """Types of digest conditions - each can have their own tree.

    Types:
        - RECEIVER: Conditions that determine who gets the digest
        - CONTENT: Conditions that determine what gets included in the digest
        - PRIORITY: Conditions that determine what is considered high priority for the digest
          - This could be used to determine sending the digest at a different time or how
            often it should be sent. It could also be used to format content.
          - Example:
            - DSRs that are due within the next week
            - Privacy requests that are due within the next week
            - Privacy requests for certain geographic regions
    """

    RECEIVER = "receiver"
    CONTENT = "content"
    PRIORITY = "priority"


class DigestCondition(ConditionalDependencyBase):
    """Digest conditional dependencies - multi-type hierarchies.

    - Multi-type hierarchy means one digest_config can have multiple independent
      condition trees, each with a different digest_condition_type (RECEIVER, CONTENT, PRIORITY)
    - Within each tree, all nodes must have the same digest_condition_type
    - This enables separate condition logic for different aspects of digest processing

    Ensures that all conditions within the same tree have the same digest_condition_type.
    This prevents logical errors where different condition types are mixed in a single
    condition tree structure.

    Example Tree Structure:
        DigestConfig (e.g., "Weekly Privacy Digest")
        ├── RECEIVER Dependency Condition Tree (who gets the digest)
        │   └── Group (AND)
        │       ├── Leaf: user.role == "admin"
        │       └── Leaf: user.department == "privacy"
        ├── CONTENT Dependency Condition Tree (what gets included)
        │   └── Group (OR)
        │       ├── Leaf: task.priority == "high"
        │       └── Leaf: task.overdue == true
        └── PRIORITY Dependency Condition Tree (when to send)
            └── Leaf: task.count >= 5
    """

    @declared_attr
    def __tablename__(cls) -> str:
        return "digest_condition"

    # We need to redefine it here so that self-referential relationships
    # can properly reference the `id` column instead of the built-in Python function.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)

    # Foreign key relationships
    digest_config_id = Column(
        String(255),
        ForeignKey("digest_config.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    parent_id = Column(
        String(255),
        ForeignKey("digest_condition.id", ondelete="CASCADE"),
        nullable=True,
        index=True,
    )

    # Digest-specific: condition category
    digest_condition_type = Column(
        EnumColumn(DigestConditionType), nullable=False, index=True
    )

    # Relationships
    digest_config = relationship("DigestConfig", back_populates="conditions")
    parent = relationship(
        "DigestCondition",
        remote_side=[id],
        back_populates="children",
        foreign_keys=[parent_id],
    )
    children = relationship(
        "DigestCondition",
        back_populates="parent",
        cascade="all, delete-orphan",
        foreign_keys=[parent_id],
    )

    # Ensure only one root condition per digest_condition_type per digest_config
    __table_args__ = (
        Index(
            "ix_digest_condition_unique_root_per_type",
            "digest_config_id",
            "digest_condition_type",
            unique=True,
            postgresql_where=text("parent_id IS NULL"),
        ),
    )

    @staticmethod
    def _validate_condition_type_consistency(db: Session, data: dict[str, Any]) -> None:
        """Validate that a condition's digest_condition_type matches its parent's type.

        Since each parent was validated when created, checking against the immediate parent
        is sufficient to ensure tree-wide consistency.

        Args:
            db: Database session for querying
            data: Dictionary containing condition data to validate (must include both parent_id and digest_condition_type)

        Raises:
            ValueError: If parent doesn't exist or digest_condition_type doesn't match parent's type
        """
        parent_id = data.get("parent_id")
        digest_condition_type = data.get("digest_condition_type")

        if not parent_id:
            # Root condition - no validation needed
            return

        # Get the parent condition
        parent = (
            db.query(DigestCondition).filter(DigestCondition.id == parent_id).first()
        )
        if not parent:
            raise ValueError(f"Parent condition with id '{parent_id}' does not exist")

        # Validate that the new condition matches the parent's digest_condition_type
        if parent.digest_condition_type != digest_condition_type:
            raise ValueError(
                f"Cannot create condition with type '{digest_condition_type}' under parent "
                f"with type '{parent.digest_condition_type}'. All conditions in the same tree "
                f"must have the same digest_condition_type."
            )

    @classmethod
    def create(
        cls,
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> "DigestCondition":
        """Create a new DigestCondition with validation."""
        # Validate condition type consistency
        cls._validate_condition_type_consistency(db, data)
        try:
            return super().create(db=db, data=data, check_name=check_name)
        except Exception as e:
            raise ConditionalDependencyError(str(e))

    def update(self, db: Session, *, data: dict[str, Any]) -> "DigestCondition":
        """Update DigestCondition with validation."""
        # Ensure validation data includes current values for fields not being updated
        validation_data = {
            "parent_id": data.get("parent_id", self.parent_id),
            "digest_condition_type": data.get(
                "digest_condition_type", self.digest_condition_type
            ),
        }

        # Validate before updating
        self._validate_condition_type_consistency(db, validation_data)
        return super().update(db=db, data=data)  # type: ignore[return-value]

    def save(self, db: Session) -> "DigestCondition":
        """Save DigestCondition with validation."""
        # Extract current object data for validation
        data = {
            "parent_id": self.parent_id,
            "digest_condition_type": self.digest_condition_type,
        }

        # Validate before saving (only if this has a parent)
        if self.parent_id:
            self._validate_condition_type_consistency(db, data)
        return super().save(db=db)  # type: ignore[return-value]

    @classmethod
    def get_root_condition(
        cls,
        db: Session,
        **kwargs: Any,
    ) -> Optional[Union[ConditionLeaf, ConditionGroup]]:
        """Get the root condition tree for a specific digest condition type.

        Implementation of the abstract base method for DigestCondition's multi-type hierarchy.
        Each digest_config can have separate condition trees for RECEIVER, CONTENT, and PRIORITY
        types. This method retrieves the root of one specific tree.

        Args:
            db: SQLAlchemy database session for querying
            **kwargs: Keyword arguments containing:
                digest_config_id: ID of the digest config
                digest_condition_type: DigestConditionType enum value
                                     Must be one of: RECEIVER, CONTENT, PRIORITY

        Returns:
            Optional[Union[ConditionLeaf, ConditionGroup]]: Root condition tree for the specified
                                                          type, or None if no conditions exist

        Raises:
            ValueError: If required parameters are missing

        Example:
            >>> # Get receiver conditions for a digest
            >>> receiver_conditions = DigestCondition.get_root_condition(
            ...     db, digest_config_id=digest_config.id,
            ...     digest_condition_type=DigestConditionType.RECEIVER
            ... )
            >>> # Get content conditions for the same digest
            >>> content_conditions = DigestCondition.get_root_condition(
            ...     db, digest_config_id=digest_config.id,
            ...     digest_condition_type=DigestConditionType.CONTENT
            ... )
        """
        digest_config_id = kwargs.get("digest_config_id")
        digest_condition_type = kwargs.get("digest_condition_type")

        if not digest_config_id or not digest_condition_type:
            raise ValueError(
                "digest_config_id and digest_condition_type are required keyword arguments"
            )

        root = (
            db.query(cls)
            .filter(
                cls.digest_config_id == digest_config_id,
                cls.digest_condition_type == digest_condition_type,
                cls.parent_id.is_(None),
            )
            .one_or_none()
        )

        if not root:
            return None

        return root.to_correct_condition_type()

    @classmethod
    def get_all_root_conditions(
        cls, db: Session, digest_config_id: str
    ) -> dict[DigestConditionType, Optional[Condition]]:
        """Get root conditions for all digest condition types"""
        return {
            condition_type: cls.get_root_condition(
                db,
                digest_config_id=digest_config_id,
                digest_condition_type=condition_type,
            )
            for condition_type in DigestConditionType
        }



"""Store SaaS Template Datasets

Revision ID: 7d82c8fc4c34
Revises: a1b2c3d4e5f6
Create Date: 2025-11-11 13:17:36.602477

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "7d82c8fc4c34"
down_revision = "a1b2c3d4e5f6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "saas_template_dataset",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("connection_type", sa.String(), nullable=False),
        sa.Column(
            "dataset_json", postgresql.JSONB(astext_type=sa.Text()), nullable=False
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_saas_template_dataset_connection_type"),
        "saas_template_dataset",
        ["connection_type"],
        unique=True,
    )
    op.create_index(
        op.f("ix_saas_template_dataset_id"),
        "saas_template_dataset",
        ["id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_saas_template_dataset_id"), table_name="saas_template_dataset"
    )
    op.drop_index(
        op.f("ix_saas_template_dataset_connection_type"),
        table_name="saas_template_dataset",
    )
    op.drop_table("saas_template_dataset")
    # ### end Alembic commands ###



import enum

from sqlalchemy import Column, Enum, ForeignKey, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base
from fides.api.models.connectionconfig import ConnectionConfig

# pylint: disable=unused-import
from fides.api.models.sql_models import System  # type: ignore[attr-defined]
from fides.config import CONFIG


class OAuthGrantType(enum.Enum):
    authorization_code = "authorization_code"
    client_credentials = "client_credentials"
    password = "password"
    implicit = "implicit"


class OAuthConfig(Base):
    """
    Stores credentials to connect fidesops to anything that uses OAuth2 for authentication.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        """Overriding base class method to set the table name."""
        return "oauth_config"

    connection_config_id = Column(
        String, ForeignKey(ConnectionConfig.id_field_path), nullable=False
    )

    grant_type = Column(Enum(OAuthGrantType), nullable=False)
    token_url = Column(String, nullable=False)
    scope = Column(String, nullable=True)
    client_id = Column(String, nullable=False)
    client_secret = Column(
        StringEncryptedType(
            String,
            CONFIG.security.app_encryption_key,
            AesGcmEngine,
            "pkcs5",
        ),
        nullable=False,
    )

    connection_config = relationship(
        "ConnectionConfig",
        back_populates="oauth_config",
    )



import json
from typing import Any, Dict, List, Optional

import requests
from loguru import logger
from oauthlib.oauth2 import BackendApplicationClient
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from fides.api.common_exceptions import ClientUnsuccessfulException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.connection_configuration import HttpsSchema
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row


class HTTPSConnector(BaseConnector[None]):
    """HTTP Connector - for connecting to second and third-party endpoints"""

    def build_uri(self) -> str:
        """
        Returns URL stored on ConnectionConfig
        """
        config = HttpsSchema(**self.configuration.secrets or {})
        return config.url

    def build_authorization_header(self) -> Dict[str, str]:
        """
        Returns Authorization headers
        """
        config = HttpsSchema(**self.configuration.secrets or {})
        return {"Authorization": config.authorization}

    def execute(
        self,
        request_body: Dict[str, Any],
        response_expected: bool,
        additional_headers: Dict[str, Any] = {},
    ) -> Optional[Dict[str, Any]]:
        """Calls a client-defined endpoint and returns the data that it responds with"""
        config = HttpsSchema(**self.configuration.secrets or {})

        def _request_with_oauth() -> requests.Response:
            """Helper function to make a request with OAuth2 authentication"""
            client_id = oauth_config.client_id
            client_secret = oauth_config.client_secret
            scopes = oauth_config.scope.split(" ") or []
            auth = HTTPBasicAuth(client_id, client_secret)
            client = BackendApplicationClient(client_id=client_id)
            session_client = OAuth2Session(client=client, scope=scopes)
            try:
                # Fetch the access token from the token URL
                session_client.fetch_token(token_url=oauth_config.token_url, auth=auth)
            except Exception as e:
                logger.error(f"Error fetching OAuth2 token: {e}")
                raise ClientUnsuccessfulException(
                    status_code=HTTP_500_INTERNAL_SERVER_ERROR
                )

            return session_client.post(
                url=config.url, headers=additional_headers, json=request_body
            )

        def _request_without_oauth() -> requests.Response:
            """Helper function to make a request without OAuth2 authentication"""
            headers = self.build_authorization_header()
            headers.update(additional_headers)
            return requests.post(url=config.url, headers=headers, json=request_body)

        oauth_config = self.configuration.oauth_config

        try:
            response = (
                _request_with_oauth()
                if oauth_config is not None
                else _request_without_oauth()
            )

            if not response_expected:
                return {}

            if not response.ok:
                logger.error("Invalid response received from webhook.")
                raise ClientUnsuccessfulException(status_code=response.status_code)

            return json.loads(response.text)

        except requests.ConnectionError:
            logger.error("HTTPS client received a connection error.")
            raise ClientUnsuccessfulException(
                status_code=HTTP_500_INTERNAL_SERVER_ERROR
            )

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Override to skip connection test
        """
        return ConnectionTestStatus.skipped

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """Return the query config that corresponds to this connector type"""
        raise NotImplementedError("Query config not implemented for HTTPS")

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """Currently not supported as webhooks are not called at the collection level"""
        raise NotImplementedError(
            "Currently not supported as webhooks are not yet called at the collection level"
        )

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """Currently not supported as webhooks are not called at the collection level"""
        raise NotImplementedError(
            "Currently not supported as webhooks are not yet called at the collection level"
        )

    def create_client(self) -> None:
        """Not required for this type"""
        return None

    def close(self) -> None:
        """Not required for this type"""



"""add classifications and user_assigned_data_categories indices

Revision ID: cd8649be3a2b
Revises: a8e0c016afd
Create Date: 2025-09-19 13:49:02.762672

"""

import sqlalchemy as sa
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
revision = "cd8649be3a2b"
down_revision = "a8e0c016afd"
branch_labels = None
depends_on = None

"""
WARNING - Conditional migration based on table size
"""


def upgrade():
    # Check stagedresource table size
    connection = op.get_bind()

    # Check stagedresource table size to decide if we should create indexes immediately
    table_size = connection.execute(
        sa.text("SELECT COUNT(*) FROM stagedresource")
    ).scalar()

    if table_size < 1000000:
        logger.info(
            f"stagedresource table has {table_size} rows, creating GIN indices directly"
        )

        # Create GIN index for user_assigned_data_categories array operations (&&, @>, <@)
        op.execute(
            "CREATE INDEX idx_stagedresource_user_categories_gin "
            "ON stagedresource USING GIN (user_assigned_data_categories)"
        )

        # Create GIN index for classifications array operations (&&, @>, <@)
        op.execute(
            "CREATE INDEX idx_stagedresource_classifications_gin "
            "ON stagedresource USING GIN (classifications)"
        )

        logger.info("GIN indices created successfully")

    else:
        logger.warning(
            "The stagedresource table has more than 1 million rows, "
            "skipping index creation. Indexes will be created during application startup "
            "via post_upgrade_index_creation.py"
        )


def downgrade():
    # Drop indices in reverse order
    logger.info("Dropping GIN indices for stagedresource arrays")

    try:
        op.drop_index("idx_stagedresource_classifications_gin", "stagedresource")
        logger.info("Dropped classifications GIN index")
    except Exception as e:
        logger.warning(f"Could not drop classifications index: {e}")

    try:
        op.drop_index("idx_stagedresource_user_categories_gin", "stagedresource")
        logger.info("Dropped user_assigned_data_categories GIN index")
    except Exception as e:
        logger.warning(f"Could not drop user_categories index: {e}")



"""adds optional duration field to assets

Revision ID: f108fa05c579
Revises: 9caf76161e55
Create Date: 2025-08-22 12:47:31.374493

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f108fa05c579"
down_revision = "9caf76161e55"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("asset", sa.Column("duration", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("asset", "duration")
    # ### end Alembic commands ###
